import os.path
from pathlib import Path
from apsimNGpy.core_utils.database_utils import get_db_table_names, read_db_table
import pandas as pd
from apsimNGpy.core.mult_cores import MultiCoreManager
from apsimNGpy.tests.unittests.base_unit_tests import BaseTester
import sys
import unittest

from apsimNGpy.settings import logger


class TestMultiCoreManager(BaseTester):
    def setUp(self):
        self.model = 'Maize'
        self.data_db = Path('test12.db').resolve()
        self.db_path_agg_func = Path('test22.db').resolve()
        self.out_path_no_agg = os.path.realpath('maize_multi_core_test_no_agg.apsimx')
        self.out_path_mean = os.path.realpath('maize_multi_core_test_mean.apsimx')
        self.jobs = ['Maize', "Soybean", 'Barley', 'Canola']
        self.core_manager_no_agg = MultiCoreManager(db_path=self.data_db, agg_func='mean')
        self.insert_data_db = os.path.realpath('test313.db')

        self.test_clear_db = os.path.realpath('test314.db')

    def tearDown(self):

        remove = [self.data_db, self.db_path_agg_func,
                  self.test_clear_db,
                  self.out_path_no_agg, self.out_path_mean]
        for element in remove:
            try:
                if os.path.exists(element): os.remove(element)

            except PermissionError:
                ...

    @staticmethod
    def create_synthetic_data():
        df = pd.DataFrame([{"cores": 2, 'sys': "linnux"}, {'sys': 'windows', 'cores': 5}])
        return df

    def test_clear_db(self):
        manager = MultiCoreManager(db_path=self.test_clear_db)
        # clear before any data is inserted
        manager.clear_db()

        # test after any data is inserted
        df = self.create_synthetic_data()
        manager.insert_data(df, table='clear_db_test')
        table_names = get_db_table_names(self.test_clear_db)
        self.assertIn('clear_db_test', table_names)

        manager.clear_db()

        if os.path.exists(self.test_clear_db):
            # then it is only cleared inside but not deleted
            table_names = get_db_table_names(self.test_clear_db)
            # if clear was successful, then the table names are empty
            self.assertFalse(table_names)
        else:
            self.assertFalse(os.path.exists(self.test_clear_db))

    def test_insert_data_db(self):
        """tests if insert_data method on MultiCoreManager is working"""
        self.insert_data_manager = MultiCoreManager(db_path=self.insert_data_db)
        # ensure files does not exists
        if os.path.exists(self.insert_data_db):
            os.remove(self.insert_data_db)
        # make up data quickly
        df = self.create_synthetic_data()
        self.insert_data_manager.insert_data(results=df, table='makeup')
        df2 = read_db_table(self.insert_data_db, 'makeup')
        self.assertFalse(df2.empty)
        self.assertEqual(df2.shape[0], df.shape[0])

    def _test_run_all_jobs(self, agg_func=None, clear_db_test=True, threads=True, cores=2):
        """tests if run_all_tests work udner various conditions"""
        core_manager_mean = MultiCoreManager(db_path=self.data_db, agg_func=agg_func)

        # using processes will fail in this environment, so threads are appropriate
        core_manager_mean.run_all_jobs(jobs=self.jobs, threads=threads, n_cores=cores, clear_db=clear_db_test,
                                       clean_up=True)
        df = core_manager_mean.results
        self.assertFalse(df.empty)
        # if mean aggregators worked, then number of rows are two
        if agg_func is not None:
            rows = df.shape[0] == len(self.jobs)

            self.assertTrue(rows, f"agg function {agg_func} failed")
        else:
            rows = df.shape[0] > len(self.jobs)
            self.assertTrue(rows)

    def test_run_all_jobs(self):
        # hard to pull off a multi_processing in this environment, we will use threads to represent all tests
        logger.info('testing run_all_tests. agg_func = mean')
        self._test_run_all_jobs(agg_func='mean', clear_db_test=True,threads=False)
        logger.info('testing run_all_tests. agg_func = None')
       # self._test_run_all_jobs(agg_func=None, clear_db_test=True)


def test_multiprocessing():
    """this is being run separately for multiprocessing in test_main"""
    if __name__ == '__main__':

        if sys.platform.startswith('win'):
            import multiprocessing as mp
            mp.freeze_support()
        unittest.main()


test_multiprocessing()



import shutil
from pathlib import Path
import os
import schedule
import time
import logging

logging.basicConfig(
    level=logging.INFO,  # or DEBUG, WARNING, ERROR, CRITICAL
    format='[%(levelname)s]: %(message)s'
)


def clean():
    path = Path("../../")
    ax = path.rglob('*.apsimx')
    db = path.rglob('*.db')
    bak = path.rglob('*.bak')
    wal = path.rglob('*.db-wal')
    remove = list(ax) + list(db) + list(path.rglob('*.db-shm')) + list(path.rglob('*.csv')) + list(bak) + list(wal)
    fl = len(remove)
    dist = path / 'dist'
    if os.path.exists(dist):
        shutil.rmtree(dist)
        print('Removed', 'dist files')
    build = path / 'build'
    if os.path.exists(build):
        shutil.rmtree(build)
        print('Removed', 'build files')
    egg = path / 'apsimNGpy.egg-info'
    if os.path.exists(egg):
        shutil.rmtree(egg)
        print('Removed', 'apsimNgpy egg file')
    for i in remove:
        try:
            i.unlink(missing_ok=True)
        except PermissionError:
            print(f"skipping '{i}' due to permission error")
    if fl > 0:
        logging.info(f" cleaned {fl} unwanted files")


if __name__ == "__main__":
    def job_with_argument(name):
        print(f"I am {name}")


    # schedule.every(1).minutes.do(clean)
    #
    # while True:
    #     schedule.run_pending()
    #     time.sleep(60)
    clean()



from dataclasses import dataclass, field
from scipy.optimize import minimize
import numpy as np
from functools import reduce, lru_cache, cache
from collections import OrderedDict
from functools import wraps
from collections import OrderedDict
from abc import ABC, abstractmethod
from wrapdisc.var import ChoiceVar, GridVar, QrandintVar, QuniformVar, RandintVar, UniformVar
import inspect
from typing import Union
import wrapdisc
from numpy import floating
from scipy.stats import norm, linregress
from typing import Union, List, Dict, Any
import pandas as pd
from abc import ABC, abstractmethod
from collections.abc import Iterable
from apsimNGpy.optimizer.optutils import compute_hyper_volume, edit_runner

ArrayLike = Union[np.ndarray, List[float], pd.Series]
from collections import OrderedDict

SING_OBJ_CONT_VAR = 'single-continuous-vars'
SING_OBJ_MIXED_VAR = 'single-mixed-vars'

@lru_cache(maxsize=None)
def get_function_param_names(func):
    """
    Return only the parameter names of a function as a tuple.
    """
    sig = inspect.signature(func)
    return tuple(sig.parameters.keys())


def _evaluate_args(*args, **kwargs):
    if not all(isinstance(arg, str) for arg in args):
        raise ValueError("all arguments must be strings")


class AbstractProblem(ABC):
    def __init__(self, apsim_model, max_cache_size=400, objectives = None, decision_vars=None, **kwargs):
        self.apsim_model = apsim_model
        self._cache = OrderedDict()
        self.max_cache_size = max_cache_size
        self._outcomes = None
        self.editor = edit_runner
        self.decision_vars = decision_vars or []
        self.objectives = objectives or []
        self.optimizer = None
        self.target_parameters = OrderedDict()

    @property
    @abstractmethod
    def optimization_type(self):
        """Must be implemented as a property in subclass"""
        pass
    def _insert_cache_result(self, *args, result):

        args = tuple(args)
        self._cache[args] = result

        if self.max_cache_size is not None and len(
                self._cache) > self.max_cache_size:  # saturation reached according to specified maxsize free space
            self._cache.popitem(last=False)

    def get_cached(self, *args):
        key = tuple(args)

        return self._cache.get(key, None)


    @cache
    def _get_editable_model(self):
        return self.apsim_model

    def clear_cache(self):
        self._cache.clear()

    @abstractmethod
    def extract_bounds(self):
        xl, xu = [], []
        for var in self.decision_vars:
            if var['v_type'] not in {'int', 'float'}:
                raise ValueError(f"Unsupported variable type: {var['v_type']}")
            xl.append(var['bounds'][0])
            xu.append(var['bounds'][1])
        print(xl, xu)
        return xl, xu

    def add_control(self, path: str, *, bounds, v_type, q=None, start_value=None, categories=None, **kwargs):
        """
        Adds a single APSIM parameter to be optimized.

        Parameters
        ----------
        path : str
            APSIM component path.

         v_type : type
            The Python type of the variable. Should be either `int` or `float` for continous variable problem or
            'uniform', 'choice',  'grid',   'categorical',   'qrandint',  'quniform' for mixed variable problem

        start_value : any (type determined by the variable type.
            The initial value to use for the parameter in optimization routines. Only required for single objective optimizations

        bounds : tuple of (float, float), optional
            Lower and upper bounds for the parameter (used in bounded optimization).
            Must be a tuple like (min, max). If None, the variable is considered unbounded or categorical or the algorithm to be used do not support bounds

        kwargs: dict
            One of the key-value pairs must contain a value of '?', indicating the parameter to be filled during optimization.
            Keyword arguments are used because most APSIM models have unique parameter structures, and this approach allows
            flexible specification of model-specific parameters. It is also possible to pass other parameters associated with the model in question to be changed on the fly.


        Returns
        -------
        self : object
            Returns self to support method chaining.

        .. warning::

            Raises a ``ValueError``
                If the provided arguments do not pass validation via `_evaluate_args`.


        .. Note::

            - This method is typically used before running optimization to define which
              parameters should be tuned.

        Example:

        .. code-block:: python

                 from apsimNGpy.core.apsim import ApsimModel
                 from apsimNGpy.core.optimizer import MultiObjectiveProblem
                 runner = ApsimModel("Maize")

                _vars = [
                {'path': '.Simulations.Simulation.Field.Fertilise at sowing', 'Amount': "?", "bounds": [50, 300],
                 "v_type": "float"},
                {'path': '.Simulations.Simulation.Field.Sow using a variable rule', 'Population': "?", 'v_type': 'float',
                 'bounds': [4, 14]}
                 ]
                problem = MultiObjectiveProblem(runner, objectives=objectives, decision_vars=_vars)
                # or
                problem = MultiObjectiveProblem(runner, objectives=objectives, None)
                problem.add_control(
                    **{'path': '.Simulations.Simulation.Field.Fertilise at sowing', 'Amount': "?", "bounds": [50, 300],
                       "v_type": "float"})
                problem.add_control(
                    **{'path': '.Simulations.Simulation.Field.Sow using a variable rule', 'Population': "?", 'v_type': 'float',
                       'bounds': [4, 14]})
        """
        self.apsim_model.check_kwargs(path, **kwargs)
        if not start_value and self.optimization_type in [SING_OBJ_CONT_VAR, SING_OBJ_MIXED_VAR]:
            raise ValueError("start_value must be provided for this problem to proceed")

        to_fill = [k for k, v in kwargs.items() if v in ('?', "")]
        if len(to_fill) != 1:
            raise ValueError("Exactly one parameter must be unspecified with '?' or '' .")

        var_spec = {
            'path': path,
            'bounds': bounds,
            'v_type': v_type,
            'q':q,
            'categories': categories,
            'start_value': start_value,
            'kwargs': kwargs
        }
        tp = to_fill[0]
        self.target_parameters[tp] = var_spec # if the parameter changes, then it is a new control
        # update decissions vars
        self.decision_vars = [self.target_parameters[i] for i in self.target_parameters]

    @abstractmethod
    def evaluate_objectives(self):
        pass



    def __str__(self):
        opt_type = self.optimization_type
        if isinstance(self.objectives, Iterable):
           ob_names = [i.__name__ for i in self.objectives ]
           ob_names = ', '.join(ob_names)
        if callable(self.objectives):
            ob_names = self.objectives.__name__
        else:
            ob_names = 'None'
        summary = f"Optimization Problem for APSIM Model: {self.model}\n"

        summary += f"===================================================\n"
        summary += f"objective_name: {ob_names}\n"
        summary += f"Number of Control Variables: {len(self.decision_vars)}\n"
        summary += "Control Variables:\n"
        for count, var in enumerate(self.decision_vars):
             count += 1
             vard = var.copy()
             vark = vard.pop('kwargs', var).copy()

             data  = vard | vark

             for key, value in data.items():
                if key in self.labels and self.outcomes:
                    x_var = getattr(self.outcomes, 'x_vars')[key]
                    summary += (
                        f"  -variable {count}: {key} = ({value}): optimized_value={x_var}\n "
                    )
                else:
                    x_var = ''
                    summary += (
                        f"  -variable {count}: {key} = ({value}):\n "
                    )
        summary += f"Results summary:: {self.outcomes}\n"
        return summary

    def __repr__(self):
        return self.__str__()
    @property
    def labels(self):
        labels = []
        for var in self.decision_vars:
            vark = var.get('kwargs', var).copy()
            data = var | vark
            for key, value in data.items():
                if value in ['?', ""]:
                 labels.append(key)
        return labels



    @property
    def indicators(self) -> List[str]:
        inds = ['mse', 'mae', 'rmse', 'rrmse', 'r2', 'wia', 'ccc', 'bias', 'slope', 'me', 'hv']
        return [i for i in inds if getattr(self, i, None) is not None]
    @property
    def outcomes(self):
        return self._outcomes

    @outcomes.setter
    def outcomes(self, value):
        self._outcomes = value
    @staticmethod
    def hv(F, reference_point=None, normalize=False, normalization_bounds=None):
        return compute_hyper_volume(F, reference_point=None, normalize=False, normalization_bounds=None)
    @staticmethod
    def rrmse(actual, predicted) -> float:
        return AbstractProblem.rmse(actual, predicted) / np.mean(actual)

    @staticmethod
    def wia(actual, predicted) -> float:
        mean_obs = np.mean(actual)
        numerator = np.sum((predicted - actual) ** 2)
        denominator = np.sum((np.abs(predicted - mean_obs) + np.abs(actual - mean_obs)) ** 2)
        return 1 - numerator / denominator

    @staticmethod
    def mse(actual, predicted) -> floating[Any]:
        return np.mean((actual - predicted) ** 2)

    @staticmethod
    def rmse(actual, predicted) -> float:
        assert len(actual) == len(predicted), f'{len(actual)} != {len(predicted)}'
        return np.sqrt(AbstractProblem.mse(actual, predicted))

    @staticmethod
    def mae(actual, predicted) -> float:
        return np.mean(np.abs(actual - predicted))

    @staticmethod
    def bias(actual, predicted) -> floating[Any]:
        return np.mean(actual - predicted)

    @staticmethod
    def slope(actual, predicted) -> float:
        slope, *_ = linregress(actual, predicted)
        return slope

    @staticmethod
    def r2(actual, predicted) -> float:
        _, _, r_value, _, _ = linregress(actual, predicted)
        return r_value ** 2

    @staticmethod
    def me(actual, predicted) -> float:
        mse_pred = np.mean((predicted - actual) ** 2)
        mse_obs = np.mean((actual - np.mean(actual)) ** 2)
        return 1 - mse_pred / mse_obs

    @staticmethod
    def rho_ci(actual, predicted, ci: str = "z-transform", conf_level: float = 0.95, na_rm: bool = False) -> Dict:
        dat = pd.DataFrame({'x': actual, 'y': predicted}).dropna() if na_rm else pd.DataFrame(
            {'x': actual, 'y': predicted})
        k = len(dat)
        r = dat['x'].corr(dat['y'])

        yb, xb = dat['y'].mean(), dat['x'].mean()
        sy2 = dat['y'].var() * (k - 1) / k
        sx2 = dat['x'].var() * (k - 1) / k
        sd1, sd2 = dat['y'].std(), dat['x'].std()

        sxy = r * np.sqrt(sx2 * sy2)
        p = 2 * sxy / (sx2 + sy2 + (yb - xb) ** 2)
        v, u = sd1 / sd2, (yb - xb) / ((sx2 * sy2) ** 0.25)
        c_b = p / r
        zv = norm.ppf(1 - (1 - conf_level) / 2)

        sep = np.sqrt(((1 - r ** 2) * p ** 2 * (1 - p ** 2) / r ** 2 +
                       (2 * p ** 3 * (1 - p) * u ** 2 / r) -
                       0.5 * p ** 4 * u ** 4 / r ** 2) / (k - 2))

        if ci == "z-transform":
            t = np.log((1 + p) / (1 - p)) / 2
            set_ = sep / (1 - p ** 2)
            llt = t - zv * set_
            ult = t + zv * set_
            llt, ult = [(np.exp(2 * val) - 1) / (np.exp(2 * val) + 1) for val in (llt, ult)]
            ci_result = pd.DataFrame({'est': [p], 'lwr.ci': [llt], 'upr.ci': [ult]})
        else:
            ci_result = pd.DataFrame({'est': [p], 'lwr.ci': [p - zv * sep], 'upr.ci': [p + zv * sep]})

        delta = dat['x'] - dat['y']
        rmean = dat.mean(axis=1)
        blalt = pd.DataFrame({'mean': rmean, 'delta': delta})

        return {'rho_c': ci_result, 's_shift': v, 'l_shift': u, 'C_b': c_b, 'blalt': blalt}

    @staticmethod
    def ccc(actual, predicted) -> float:
        return AbstractProblem.rho_ci(actual, predicted)['rho_c']['est'].iloc[0]


@dataclass(slots=True, frozen=True)
class VarDesc:
    model_type: str
    model_name: str
    parameter_name: str
    vtype: Union[str, object, wrapdisc.var]
    label: str
    start_value: Union[float, int]
    bounds: tuple[float, int]


if __name__ == '__main__':
    pred_data = np.array([
        8469.616,
        4668.505,
        555.047,
        3504.000,
        7820.075,
        8823.517,
        3587.101,
        2939.152,
        8379.435,
        7370.301
    ])

    obs = np.array([
        7000.0,
        5000.505,
        1000.047,
        3504.000,
        7820.075,
        7000.517,
        3587.101,
        4000.152,
        8379.435,
        4000.301
    ])
    val = dict(mse=AbstractProblem.mse(obs, pred_data),
               rmse=AbstractProblem.rmse(obs, pred_data),
               mae=AbstractProblem.mae(obs, pred_data),
               rrmse=AbstractProblem.rrmse(obs, pred_data),
               wia=AbstractProblem.wia(obs, pred_data),
               ccc=AbstractProblem.ccc(obs, pred_data),
               slope=AbstractProblem.slope(obs, pred_data),
               r2=AbstractProblem.r2(obs, pred_data))
    import pprint
    pprint.pprint(val, indent=4, width=4)



from apsimNGpy.core.apsim import ApsimModel

model = ApsimModel('Maize')
out_cultivar = 'B_110-e'
com_path = '[Phenology].Juvenile.Target.FixedValue'
model.edit_model(model_type='Cultivar', model_name='B_110', new_cultivar_name=out_cultivar,
                 commands=com_path, values=16, verbose=True,
                 cultivar_manager='Sow using a variable rule')
model.run()
print(model.results.Yield.mean())
pc = model.inspect_model_parameters(model_type='Cultivar', model_name=out_cultivar)
print(pc[com_path])

model.edit_model(model_type='Cultivar', model_name='B_110-e', new_cultivar_name=out_cultivar,
                 commands=com_path, values=406, verbose=True,
                 cultivar_manager='Sow using a variable rule')
pc = model.inspect_model_parameters(model_type='Cultivar', model_name=out_cultivar)
model.run()
print(model.results.Yield.mean())
print(pc[com_path])




class ApsimNGpyError(Exception):
    """Base class for all apsimNGpy-related exceptions. These errors are more descriptive than just rising a value error"""
    pass


class InvalidInputErrors(ApsimNGpyError):
    """Raised when the input provided is invalid or improperly formatted."""
    pass


class ForgotToRunError(ApsimNGpyError):
    """Raised when a required APSIM model run was skipped or forgotten."""
    pass


class EmptyDateFrameError(ApsimNGpyError):
    """Raised when a DataFrame is unexpectedly empty."""
    pass


class NodeNotFoundError(ApsimNGpyError):
    """Raised when a specified model node cannot be found."""
    pass


class ModelNotFoundError(ApsimNGpyError):
    """Raised when a specified model  cannot be found."""
    pass


class CastCompilationError(ApsimNGpyError):
    """Raised when the C# cast helper DLL fails to compile."""
    pass


class ApsimNotFoundError(ApsimNGpyError):
    """Raised when the APSIM executable or directory is not found."""
    pass


class ApsimBinPathConfigError(ApsimNGpyError):
    """Raised when the APSIM bin path is misconfigured or incomplete."""
    pass


class TableNotFoundError(ApsimNGpyError):
    """table not found error."""
    pass



"""
Evaluate predicted vs. observed data using statistical and mathematical metrics.
For detailed metric definitions, see Archontoulis et al. (2015).
"""

__all__ = ['Validate']

import numpy as np
import pandas as pd
from attr import dataclass
from scipy.stats import norm, linregress
from typing import Union, List, Dict


ArrayLike = Union[np.ndarray, List[float], pd.Series]


@dataclass
class Validate:
    """
    Compares predicted and observed values using various statistical metrics.
    """

    actual: ArrayLike
    predicted: ArrayLike

    METRICS = ['RMSE', 'MAE', 'MSE', 'RRMSE', 'bias', 'ME', 'WIA', 'R2', 'CCC', 'slope']

    def __post_init__(self):
        self.actual = np.asarray(self.actual)
        self.predicted = np.asarray(self.predicted)
        self.obs_mean = np.mean(self.actual)
        self.pred_mean = np.mean(self.predicted)
        self.std_actual = np.std(self.actual)
        self.std_predicted = np.std(self.predicted)
        assert len(self.actual) == len(self.predicted), "Predicted and actual arrays must have the same length."

    def RRMSE(self) -> float:
        rmse = self.RMSE()
        return rmse / np.mean(self.actual)

    def WIA(self) -> float:
        mean_obs = np.mean(self.actual)
        numerator = np.sum((self.predicted - self.actual) ** 2)
        denominator = np.sum((np.abs(self.predicted - mean_obs) + np.abs(self.actual - mean_obs)) ** 2)
        return 1 - numerator / denominator

    def MSE(self) -> float:
        return np.mean((self.actual - self.predicted) ** 2)

    def RMSE(self) -> float:
        return np.sqrt(self.MSE())

    def MAE(self) -> float:
        return np.mean(np.abs(self.actual - self.predicted))

    def bias(self) -> float:
        return np.mean(self.actual - self.predicted)

    def slope(self) -> float:
        slope, *_ = linregress(self.actual, self.predicted)
        return slope

    def R2(self) -> float:
        _, _, r_value, _, _ = linregress(self.actual, self.predicted)
        return r_value ** 2

    def ME(self) -> float:
        mse_pred = np.mean((self.predicted - self.actual) ** 2)
        mse_obs = np.mean((self.actual - np.mean(self.actual)) ** 2)
        return 1 - mse_pred / mse_obs

    def CCC(self) -> float:
        return self.rho_ci()['rho_c']['est'].iloc[0]

    def rho_ci(self, ci: str = "z-transform", conf_level: float = 0.95, na_rm: bool = False) -> Dict:
        dat = pd.DataFrame({'x': self.actual, 'y': self.predicted}).dropna() if na_rm else pd.DataFrame({'x': self.actual, 'y': self.predicted})
        k = len(dat)
        r = dat['x'].corr(dat['y'])

        yb, xb = dat['y'].mean(), dat['x'].mean()
        sy2 = dat['y'].var() * (k - 1) / k
        sx2 = dat['x'].var() * (k - 1) / k
        sd1, sd2 = dat['y'].std(), dat['x'].std()

        sxy = r * np.sqrt(sx2 * sy2)
        p = 2 * sxy / (sx2 + sy2 + (yb - xb) ** 2)
        v, u = sd1 / sd2, (yb - xb) / ((sx2 * sy2) ** 0.25)
        C_b = p / r
        zv = norm.ppf(1 - (1 - conf_level) / 2)

        sep = np.sqrt(((1 - r ** 2) * p ** 2 * (1 - p ** 2) / r ** 2 +
                       (2 * p ** 3 * (1 - p) * u ** 2 / r) -
                       0.5 * p ** 4 * u ** 4 / r ** 2) / (k - 2))

        if ci == "z-transform":
            t = np.log((1 + p) / (1 - p)) / 2
            set_ = sep / (1 - p ** 2)
            llt = t - zv * set_
            ult = t + zv * set_
            llt, ult = [(np.exp(2 * val) - 1) / (np.exp(2 * val) + 1) for val in (llt, ult)]
            ci_result = pd.DataFrame({'est': [p], 'lwr.ci': [llt], 'upr.ci': [ult]})
        else:
            ci_result = pd.DataFrame({'est': [p], 'lwr.ci': [p - zv * sep], 'upr.ci': [p + zv * sep]})

        delta = dat['x'] - dat['y']
        rmean = dat.mean(axis=1)
        blalt = pd.DataFrame({'mean': rmean, 'delta': delta})

        return {'rho_c': ci_result, 's_shift': v, 'l_shift': u, 'C_b': C_b, 'blalt': blalt}

    def evaluate(self, metric: str = 'RMSE') -> float:
        assert metric in self.METRICS, f"Metric '{metric}' is not supported."
        return getattr(self, metric)()

    def evaluate_all(self, verbose: bool = False) -> Dict[str, float]:
        results = {metric: getattr(self, metric)() for metric in self.METRICS}
        if verbose:
            for k, v in results.items():
                print(f"{k}: {v:.4f}")
        return results


if __name__ == "__main__":
    x_data = np.array([1.2, 2.4, 3.6, 4.8, 5.0])
    y_data = np.array([2.0, 3.5, 4.2, 5.7, 6.0])

    validator = Validate(x_data, y_data)
    all_metrics = validator.evaluate_all(verbose=True)




if True:

    # running factorial experiments
    import os

    from apsimNGpy.core.experimentmanager import ExperimentManager as Experiment
    if __name__ == '__main__':
        experiment = Experiment("Maize", )
        experiment.init_experiment(permutation=True)
        experiment.add_factor("[Fertilise at sowing].Script.Amount = 0, 300")
        experiment.add_factor(specification="[Sow using a variable rule].Script.CultivarName = \
                           Dekalb_XL82, Melkassa, B_110, Pioneer_34K77",
                              )
        experiment.run()
        df = experiment.results
        g = experiment.cat_plot(x='Amount', y="Yield", kind='bar', aspect=1, height=8, hue='CultivarName')
        experiment.render_plot(ylabel= r'Corn grain yield (kg ha $^{-1}$)',
                               xlabel= 'Nitrogen fertilizers (kg ha $^{-1}$)',
                               save_as="Maize_experiment.png", show=True)








        os.startfile("Maize_experiment.png")
        plt.close()
        df.groupby(['Amount', 'CultivarName'])['Yield'].mean()




from apsimNGpy.core.config import get_apsim_bin_path
from apsimNGpy.core.cs_resources import start_pythonnet

# from apsimNGpy.core_utils.cs_utils import CastHelper
__all__ = ['start_pythonnet']

from apsimNGpy.core.config import (
    get_apsim_bin_path,
    set_apsim_bin_path,
    auto_detect_apsim_bin_path
)
from apsimNGpy.core_utils.utils import timer
from apsimNGpy.exceptions import (
    InvalidInputErrors,
    ForgotToRunError,
    EmptyDateFrameError
)

__version__ = '0.39.8.14'

__all__.extend([
    'InvalidInputErrors',
    'ForgotToRunError',
    'EmptyDateFrameError',
    'get_apsim_bin_path',
    'auto_detect_apsim_bin_path',
    'set_apsim_bin_path',
    'timer',
    '__version__',
    'version'
])

# Conditionally import core modules that rely on APSIM binary being available
if get_apsim_bin_path():

    from apsimNGpy.core import core, apsim, base_data
    from apsimNGpy.core.apsim import ApsimModel
    from apsimNGpy.core.mult_cores import MultiCoreManager

    from apsimNGpy.validation import evaluator, eval_methods
    from apsimNGpy.parallel.process import (
        custom_parallel,
        run_apsimx_files_in_parallel
    )
    from apsimNGpy.core_utils import database_utils
    from apsimNGpy.manager.weathermanager import (
        get_met_from_day_met,
        get_met_nasa_power,
        get_iem_by_station,
        get_weather
    )
    from apsimNGpy.optimizer.single import ContinuousVariable, MixedVariable

    try:
        from apsimNGpy.optimizer.moo import MultiObjectiveProblem
    except ImportError:
        MultiObjectiveProblem = None
    from apsimNGpy.core.pythonet_config import is_file_format_modified
    if is_file_format_modified():
        from apsimNGpy.core.experimentmanager import ExperimentManager

    __all__.extend([
        'MultiCoreManager',
        'core',
        'apsim',
        'base_data',
        'ApsimModel',
        'evaluator',
        'eval_methods',
        'custom_parallel',
        'run_apsimx_files_in_parallel',
        'database_utils',
        'get_met_from_day_met',
        'get_met_nasa_power',
        'get_iem_by_station',
        'get_weather',
        "ContinuousVariable",
        "MixedVariable",
        'MultiObjectiveProblem'
    ])

if is_file_format_modified():
    __all__.append(
     'ExperimentManager'
    )


import os
import unittest
from pathlib import Path
from typing import Set

import pandas as pd

from apsimNGpy.core.config import apsim_version, stamp_name_with_version
from apsimNGpy.core.core import CoreModel, Models
from apsimNGpy.core.model_tools import find_model, validate_model_obj, find_child
from apsimNGpy.tests.unittests.base_unit_tests import BaseTester
from apsimNGpy.core.base_data import load_default_simulations
from apsimNGpy.core.pythonet_config import is_file_format_modified

IS_NEW_APSIM = is_file_format_modified()
RUN_ONLINE = os.getenv("APSIM_ONLINE_TESTS") == "1"


class TestCoreModel(BaseTester):
    """Unit tests for CoreModel and related helper utilities."""

    @classmethod
    def setUpClass(cls):
        # Use a temporary working area per test class
        cls._tmpdir = Path.cwd() / "test_core_tmp"
        cls._tmpdir.mkdir(parents=True, exist_ok=True)

        # Avoid mutating global CWD for the process; keep a reference
        cls._old_cwd = Path.cwd()
        os.chdir(cls._tmpdir)

        cls.VERSION = apsim_version()

    @classmethod
    def tearDownClass(cls):
        # Restore original CWD
        os.chdir(cls._old_cwd)

        # Best-effort cleanup of temp folder if empty
        try:
            cls._tmpdir.rmdir()
        except OSError:
            # Directory not empty or in use; leave it
            pass

    def setUp(self):
        # Track every path we create so tearDown can clean it
        self.paths: Set[Path] = set()

        # Precompute filenames (stamped)
        self.user_file_name = Path(
            stamp_name_with_version(os.path.realpath("test_core_model.apsimx"))
        )
        self.weather_file_nasa = Path(os.path.realpath("test_met_nasa.met"))
        self.weather_file_daymet = Path(os.path.realpath("test_met_daymet.met"))
        self.test_user_out_path = os.path.realpath("test_user_file_name.apsimx")
        self.test_user_out_path_stamped = stamp_name_with_version(self.test_user_out_path)


        self.model = "Maize"
        self.test_ap_sim = CoreModel(self.model, out_path=str(self.user_file_name))

        # Register files for cleanup (some may be created later)
        self.paths.update(
            {
                self.user_file_name,
                self.weather_file_nasa,
                self.weather_file_daymet,
                self.test_user_out_path,
            }
        )

    def tearDown(self):
        # Clean model artifacts
        try:
            self.test_ap_sim.clean_up()
        except Exception:
            pass

        # Delete any files we created
        for p in list(self.paths):
            try:
                if Path(p).exists():
                    os.remove(p)
            except Exception:
                # Ignore stubborn files; test should still conclude
                pass

    # ---------------- Core path/output tests ----------------

    def test_out_path(self):
        """User-specified output path is respected and non-empty."""
        if os.path.exists(self.test_user_out_path_stamped):
            os.remove(self.test_user_out_path_stamped)

        model = CoreModel(self.model, out_path=str(self.test_user_out_path))
        print(model.path, 'from model')
        print(self.test_user_out_path)
        self.paths.add(Path(model.path))

        self.assertEqual(str(model.path), str(self.test_user_out_path_stamped))
        self.assertGreater(Path(self.test_user_out_path_stamped).stat().st_size, 0, "out_path is empty")

        model.clean_up()

    def test_random_out_path(self):
        """Random output path is created when not provided."""
        model = CoreModel(self.model, out_path=None)
        self.paths.add(Path(model.path))
        self.assertTrue(Path(model.path).exists())
        model.clean_up()

    # ---------------- Running & results ----------------

    def test_run(self):
        """Run with list and string report names; confirm DataFrame results."""
        self.test_ap_sim.run(report_name=["Report"])
        self.assertIsInstance(
            self.test_ap_sim.results, pd.DataFrame, "Expected DataFrame with list report_name"
        )

        self.test_ap_sim.run(report_name="Report", verbose=False)
        df = self.test_ap_sim.results
        self.assertIsInstance(df, pd.DataFrame)
        self.assertFalse(df.empty, "Model ran, but results are empty")
        self.assertTrue(self.test_ap_sim.ran_ok, f"Run failed: {repr(self.test_ap_sim)}")

        self.paths.add(self.user_file_name)

    def test_get_simulated_output(self):
        """get_simulated_output returns a DataFrame after run."""
        self.test_ap_sim.run()
        if not self.test_ap_sim.ran_ok:
            self.skipTest("Model did not run successfully; skipping output test.")
        repos = self.test_ap_sim.get_simulated_output(report_names="Report")
        self.assertIsInstance(repos, pd.DataFrame, f"Expected DataFrame, got {type(repos)}")

    # ---------------- Inspection helpers ----------------

    def test_get_reports(self):
        self.assertIsInstance(self.test_ap_sim.inspect_model("Report"), list)

    def test_inspect_clock(self):
        self.assertIsInstance(self.test_ap_sim.inspect_model("Clock"), list)

    def test_inspect_simulation(self):
        self.assertIsInstance(
            self.test_ap_sim.inspect_model("Models.Core.Simulation", fullpath=False), list
        )

    def test_find_simulations(self):
        """find_simulations works with None, str, tuple, and list."""
        sim = "Simulation"
        msg = f"find_simulations failed to return requested simulation object: {sim}"

        self.assertTrue(self.test_ap_sim.find_simulations(), msg=msg)
        self.assertTrue(self.test_ap_sim.find_simulations(simulations=sim), msg=msg)
        self.assertTrue(self.test_ap_sim.find_simulations(simulations=(sim,)), msg=msg)
        self.assertTrue(self.test_ap_sim.find_simulations(simulations=[sim]), msg=msg)

    # ---------------- Management & editing ----------------

    def test_update_mgt(self):
        """update_mgt accepts a management dict/list and applies changes."""
        amount = 23.557777
        script_name = "Fertilise at sowing"

        # If your API expects a dict:
        mgt_script = {"Name": script_name, "Amount": amount}
        # If it expects a list of dicts, use: [mgt_script]

        self.test_ap_sim.update_mgt(management=mgt_script)

        value = self.test_ap_sim.inspect_model_parameters(
            "Manager", simulations="Simulation", model_name=script_name, parameters="Amount"
        )
        amount_in = float(value["Amount"])
        self.assertEqual(amount_in, amount)

    def test_replace_soil_property_values(self):
        parameter = "Carbon"
        param_values = [2.4, 1.4]

        self.test_ap_sim.replace_soil_property_values(
            parameter=parameter, param_values=param_values, soil_child="Organic"
        )

        lst = self.test_ap_sim.inspect_model_parameters(
            model_type="Organic", simulations="all", model_name="Organic", parameters="Carbon"
        )["Carbon"].tolist()

        self.assertIsInstance(lst, list, f"Expected list, got {type(lst)}")
        self.assertTrue(lst)
        self.assertEqual(
            lst[:2],
            param_values,
            f"replace_soil_property_values failed: got {lst[:2]} vs {param_values}",
        )

    def test_replace_soil_properties_by_path(self):
        param_values = [1.45, 1.95]
        self.test_ap_sim.edit_model(model_type="Organic", model_name="Organic", Carbon=param_values)

    def test_update_mgt_by_path(self):
        # fmt='.'
        path_dot = ".Simulations.Simulation.Field.Sow using a variable rule"
        self.test_ap_sim.update_mgt_by_path(path=path_dot, Population=7.5, fmt=".")
        self.test_ap_sim.run()
        self.assertTrue(self.test_ap_sim.ran_ok, "Simulation did not run for fmt='.'")

        # fmt='/'
        path_slash = "/Simulations/Simulation/Field.Sow using a variable rule"
        self.test_ap_sim.update_mgt_by_path(path=path_slash, Population=7.5, fmt="/")
        self.test_ap_sim.run()
        self.assertTrue(self.test_ap_sim.ran_ok, "Simulation did not run for fmt='/'")

    def test_create_experiment(self):
        """Creates a factorial experiment, adds a factor, and runs."""
        if IS_NEW_APSIM:
            self.skipTest("Experiment path tested elsewhere for new APSIM format.")
        else:
            self.test_ap_sim.create_experiment()
            self.test_ap_sim.add_factor(
                specification="[Fertilise at sowing].Script.Amount = 0 to 200 step 20",
                factor_name="Nitrogen",
            )
            self.test_ap_sim.run()
            self.assertTrue(self.test_ap_sim.ran_ok, "Run failed after adding experiment/factor")

    def test_add_crop_replacements(self):
        self.test_ap_sim.add_crop_replacements(_crop="Maize")
        xp = find_child(self.test_ap_sim.Simulations, child_name="Replacements", child_class="Models.Core.Folder")
        self.assertTrue(bool(xp), "Adding replacements failed")

    def test_replace_soils_values_by_path(self):
        node = ".Simulations.Simulation.Field.Soil.Organic"
        self.test_ap_sim.replace_soils_values_by_path(node_path=node, indices=[0], Carbon=1.3)
        v = self.test_ap_sim.get_soil_values_by_path(node, "Carbon")["Carbon"][0]
        self.assertEqual(v, 1.3)

    def test_add_db_table(self):
        self.test_ap_sim.add_db_table(
            variable_spec=[
                "[Clock].Today",
                "[Soil].Nutrient.TotalC[1]/1000 as SOC1",
                "[Maize].Grain.Total.Wt*10 as Yield",
            ],
            rename="report4",
            set_event_names=["[Maize].Harvesting", "[Clock].EndOfYear"],
        )

        reports = self.test_ap_sim.inspect_model("Models.Report()", fullpath=False)
        self.assertIn("report4", reports, "report4 not found in reports")

        self.test_ap_sim.run("report4")
        self.assertTrue(self.test_ap_sim.ran_ok, "Run failed after adding report table.")

    def test_loading_defaults_with(self):
        model = CoreModel(model="Maize")
        model.run()
        self.assertTrue(model.ran_ok, "Run failed after loading defaults in CoreModel.")
        self.paths.add(Path(model.path))
        model.clean_up()

    def test_edit_with_path_som(self):
        model = CoreModel(model="Maize")
        initial_ratio = 200.0
        model.edit_model_by_path(
            ".Simulations.Simulation.Field.SurfaceOrganicMatter",
            InitialCNR=initial_ratio,
            verbose=False,
        )
        icnr = model.inspect_model_parameters(
            model_type="SurfaceOrganicMatter",
            model_name="SurfaceOrganicMatter",
            parameters="InitialCNR",
        )
        self.assertEqual(
            icnr["InitialCNR"], initial_ratio, "InitialCNR not updated by edit_model_by_path"
        )
        self.paths.add(Path(model.path))
        model.clean_up()

    # ---------------- Model lookup utilities ----------------

    def test_find_models_and_validate(self):
        """find_model/validate_model_obj should resolve names & paths correctly."""
        clock_model = find_model("clock")
        validate_model_obj(clock_model)
        self.assertEqual(clock_model, Models.Clock)

        simulation_model = find_model("Simulation")
        validate_model_obj(simulation_model)
        self.assertEqual(simulation_model, Models.Core.Simulation)

        clock = validate_model_obj("Models.Clock")
        self.assertEqual(clock, Models.Clock)

        simulation = validate_model_obj("Models.Core.Simulation")
        self.assertEqual(simulation, Models.Core.Simulation)

        experiment = validate_model_obj("Experiment")
        self.assertEqual(experiment, Models.Factorial.Experiment)

        factors = validate_model_obj("Factors")
        self.assertEqual(factors, Models.Factorial.Factors)

    def test_add_model_simulation(self):
        if IS_NEW_APSIM:
            self.skipTest("Add model simulation test not applicable for new APSIM format.")
        else:
            try:
                # NOTE: keep the kwarg name consistent with your API (`adoptive_parent` vs `adaptive_parent`)
                self.test_ap_sim.add_model(
                    "Simulation",
                    adoptive_parent="Simulations",
                    rename="soybean_replaced",
                    source="Soybean",
                    override=True,
                )
                sims = self.test_ap_sim.inspect_model("Simulation", fullpath=False)
                self.assertIn("soybean_replaced", sims, "Adding simulation failed")
            finally:
                self.test_ap_sim.remove_model(model_class="Simulation", model_name="soybean_replaced")

    def test_inspect_file(self):
        inspected = self.test_ap_sim.inspect_file(console=False)
        self.assertTrue(inspected)

    # ---------------- Weather (network-dependent; optional) ----------------

    def test_get_weather_from_web_nasa(self):
        model = load_default_simulations("Maize")
        model.get_weather_from_web(
            lonlat=(-93.50456, 42.601247),
            start=1990,
            end=2001,
            source="nasa",
            filename=str(self.weather_file_nasa),
        )
        model.run()
        self.assertTrue(model.ran_ok)
        self.paths.add(Path(model.path))
        model.clean_up()

    def test_get_weather_from_web_daymet(self):
        model = load_default_simulations("Maize")
        model.get_weather_from_web(
            lonlat=(-93.50456, 42.601247),
            start=1990,
            end=2001,
            source="daymet",
            filename=str(self.weather_file_daymet),
        )
        model.run()
        self.assertTrue(model.ran_ok)
        self.paths.add(Path(model.path))
        model.clean_up()


if __name__ == "__main__":
    unittest.main()



import os.path
import shutil
import unittest
import os
# Import the module where CoreModel class is defined
from apsimNGpy.core.config import (set_apsim_bin_path, get_apsim_bin_path, apsim_version, load_crop_from_disk,
                                   auto_detect_apsim_bin_path, get_bin_use_history)
from apsimNGpy.tests.unittests.base_unit_tests import BaseTester, path
from pathlib import Path

TEST_PATH = "/path/to/test/bin"  # apsim bin path to test

auto = auto_detect_apsim_bin_path()


class TestConfig(BaseTester):
    def setUp(self):
        self.model = 'Maize'
        self.user_file_name = os.path.realpath('user_file_name.apsimx')
        self.work_space = 'config_ws'
        Path(self.work_space).mkdir(exist_ok=True)
        self.paths = set()  # for storing random paths

    def test_get_apsim_bin_path(self):
        abp = get_apsim_bin_path()
        p_true = path.exists(abp)
        if not abp:
            self.skipTest('Could not find apsim bin. perhaps not yet set')
        self.assertTrue(p_true)

    def test_get_apsim_version(self):
        import re
        version = apsim_version()
        pattern = r"^APSIM \d{4}\.\d+\.\d+\.\d+$"
        match = re.match(pattern, version)
        self.assertTrue(match, 'Failed to detect Current APSIM version')

    def test_get_bin_use_history(self):
        """test if apsimNGpy is recording the used binary path.

         Please remember only those stills existing on the computer are recorded"""
        bin_history = get_bin_use_history()
        if not bin_history:
            self.skipTest('No binary history available, skipping test..')
        self.assertIsInstance(bin_history, list, msg='Binary history is not being detected')
        # ensure all exists
        all_exists = all([os.path.exists(i) for i in bin_history])
        self.assertTrue(all_exists, msg='some path are no longer available on the computer')

    def test_load_crop_from_disk_user_name(self):
        """Test user supplied path"""
        if os.path.exists(self.user_file_name):
            os.remove(self.user_file_name)
        crop = load_crop_from_disk(self.model, out=self.user_file_name)
        self.assertTrue(crop)
        self.assertGreater(os.path.getsize(self.user_file_name), 0,
                           f'failed to load default model: using {self.user_file_name} ')

    def test_load_crop_from_disk_default_random_name(self):
        """test if random name is successful"""
        crop = load_crop_from_disk(self.model)
        self.assertGreater(os.path.getsize(crop), 0, msg='failed to load using random name')
        self.paths.add(crop)

    def test_work_space_load_crop_from_disk(self):
        """test work space provided actually works"""
        # load and return crop path
        crop = load_crop_from_disk(self.model, work_space=self.work_space)
        dir_name = os.path.dirname(crop)
        self.assertEqual(dir_name, self.work_space)
        # also check that the file has been created
        self.assertGreater(os.path.getsize(crop), 0,
                           msg=f'failed to load using random name and provided work_space {self.work_space}')
        self.paths.add(crop)

    def test_set_apsim_bin_path(self):
        """
        Test that set_apsim_bin_path
        This will always fail if the TEST_PATH above is not pointing correctly to the correct path of APSIM bin files
        @return:
        """
        # Initially retrieve the current APSIM binary path
        original_path = get_apsim_bin_path()

        try:
            # Set a new test path

            set_apsim_bin_path(TEST_PATH, raise_errors=False)

            # Retrieve the path after setting it
            new_path = get_apsim_bin_path()

            # Check if the new path is as expected
            self.assertEqual(new_path, new_path, "The APSIM binary path was not updated correctly.")
        # self.skipTest('The TEST_PATH was not updated')

        finally:
            # Clean up: restore an original path to avoid side effects in other tests
            # set_apsim_bin_path(original_path)
            pass

    def tearDown(self):
        if os.path.exists(self.work_space):
            shutil.rmtree(self.work_space)
        if os.path.exists(self.user_file_name):
            os.remove(self.user_file_name)

        for pa in self.paths:
            if os.path.exists(pa):
                os.remove(pa)


if __name__ == '__main__':
    unittest.main()



"""
This is an internal script to automate the tests. Therefore, some packages used are not added to apsimNGpy project requirements
Wakes up every specified period of time to run the test after checking if there is a new version
"""
import gc
import os
import sys
import time
from pathlib import Path
import pandas as pd
import subprocess
import sys
import subprocess

try:
    import bs4
except (ModuleNotFoundError, ImportError):
    print("🔄 Attempting to install beautifulsoup4...")
    try:
        subprocess.check_call(
            [sys.executable, "-m", "pip", "install", "beautifulsoup4"],

        )
        import bs4
        print("✅ Successfully installed beautifulsoup4.")
    except subprocess.CalledProcessError as e:
        print("❌ Pip installation failed!")
        print("Exit code:", e.returncode)
        print("Command:", e.cmd)
        sys.exit(1)  # Or raise if you want the script to stop


from pandas import DataFrame
from pywinauto import Application, keyboard
import requests
from apsimNGpy.core.config import set_apsim_bin_path
from apsimNGpy.settings import logger
from apsimNGpy.tests.tester_main import run_suite

home = Path.home().joinpath("AppData", 'Local', 'Programs')
csv_path = home.joinpath("test_logs.csv")

user_email = 'magalarich20@gmail.com'

df = pd.read_csv(csv_path)
from datetime import datetime

date_STR = datetime.now().strftime("%y-%m-%d-%H-%M-%S")
date_now = datetime.now()
# Constants
email = user_email
CURRENT_VERSION = "2025.07.7820.0"


RUNS = 0
def v_next(previous):
    _, _, sub_version, _ = previous.split('.')
    sub = int(sub_version)
    sub += 1
    return sub


def insert_data(file_name):
    file_name = str(file_name)
    if not os.path.exists(csv_path):
        DataFrame().to_csv(csv_path)
    date_str = datetime.now().strftime("%y-%m-%d-%H-%M-%S")
    df = pd.read_csv(csv_path)
    installer = df.get('installer', [])
    if installer is not None and not isinstance(installer, list):
        installer = installer.tolist()
    file_without_suffix = file_name.removesuffix('.exe')
    if not file_without_suffix in installer:

        installer.append(file_without_suffix)
        date_ti = df.get('date_time', [])
        if date_ti is not None and not isinstance(date_ti, list):
            date_ti = date_ti.tolist()
        date_ti.append(date_str)
        df = pd.DataFrame({'installer': installer, 'date_time': date_ti})
        logger.info('saving data succeeded')
        df.to_csv(csv_path, index=False)
        del df
        gc.collect()


def get_last_installed():
    installer = pd.read_csv(csv_path)['installer'].tolist()
    if installer is not None:
        return installer[-1]


def extract_version_from_url(url):
    from urllib.parse import urlparse, parse_qs
    # Parse the URL query
    query = urlparse(url).query
    params = parse_qs(query)

    version = params.get("version", [None])[0]
    return version


def get_next_version(last_version: str) -> str:
    try:
        parts = last_version.strip().split(".")
        if len(parts) != 4:
            raise ValueError("Invalid version format")

        # Convert last BUILD part to int and increment
        build = int(parts[2])
        revision = int(parts[3])  # usually 0, but handled anyway

        new_version = f"{parts[0]}.{parts[1]}.{build + 1}.0"
        return new_version

    except Exception as e:
        raise RuntimeError(f"Failed to compute next version from {last_version}: {e}")


def notify(NEXTVERSION):
    file_name = f"{NEXTVERSION}.exe"
    if os.path.exists(file_name):
        logger.info(f"skipping download finishing previously downloaded version ({file_name})")
        return True


def download_installer():
    global RUNS
    try:
        LAST = get_last_installed()
        # logger.info("last apsim version is {}".format(LAST))
        NEXTVERSION = get_next_version(LAST)

        session = requests.Session()
        headers = {
            "User-Agent": "Mozilla/5.0"
        }

        # Step 1: Get the registration page to extract the token
        resp = session.get(
            "https://registration.apsim.info/?product=APSIM%20Next%20Generation&version",
            headers=headers
        )
        soup = BeautifulSoup(resp.text, "html.parser")

        token_input = soup.find("input", {"name": "__RequestVerificationToken"})
        token = token_input["value"]

        # Step 2: Submit the form with email and token
        post_data = {
            "email": email,
            "__RequestVerificationToken": token,
            "Sender": "LandingPage"
        }
        resp2 = session.post(
            "https://registration.apsim.info/?product=APSIM%20Next%20Generation&version",
            data=post_data,
            headers=headers
        )

        # Step 3: Parse the returned HTML for download links
        soup2 = BeautifulSoup(resp2.text, "html.parser")
        links = soup2.find_all("a")

        # Step 4: Find the correct version
        download_link = None
        vp = v_next(previous=LAST)
        for link in links:
            href = link.get("href", "")
            if f'{vp}' in href:
                print(href)
                next_v = extract_version_from_url(href)
                nv = notify(next_v)
                file_name =f'{next_v}.exe'
                if nv:
                    return file_name

                download_link = "https://registration.apsim.info" + href
                break
            else:
                #logger.info(f'Could not find APSIM version sub number: {vp}. perhaps no yet published')
                return 0

        print(f"✅ Found link: {download_link}")

        # Step 5: Use requests to get the installer
        resp3 = session.get(
            f"https://registration.apsim.info/?product=APSIM%20Next%20Generation&version={next_v}&platform=Windows",
            headers=headers,
            stream=True
        )

        with open(file_name, "wb") as f:
            for chunk in resp3.iter_content(chunk_size=8192):
                if chunk:
                    f.write(chunk)

        logger.info(f"✅ Downloaded and saved as: {file_name}")
        return file_name
    finally:
        RUNS +=1
        print(f'NO of attempts: {RUNS}')


def get_installed_path(file_path):
    file = str(file_path).removesuffix('.exe')
    six = file[-6:]
    installed_path = list(home.rglob(f"*{six}"))
    if installed_path:
        return installed_path[0]
    else:
        raise FileNotFoundError(f"no installed path for {file_path}")


def main():
    try:
        installer_name = download_installer()
        if not installer_name:
            return None
        logger.info(f'starting apsim{installer_name} installer')
        app = Application(backend='uia')
        app.start(cmd_line=installer_name)
        time.sleep(1)
        keyboard.send_keys("{DOWN}")
        keyboard.send_keys("{SPACE}")
        logger.info('selecting install path')
        time.sleep(1)
        keyboard.send_keys("{ENTER}")

        logger.info('selecting next')
        time.sleep(2)
        keyboard.send_keys("{ENTER}")
        logger.info('selecting apsimx icon or desktop')
        time.sleep(2)
        keyboard.send_keys("{SPACE}")
        time.sleep(2)
        keyboard.send_keys("{ENTER}")
        logger.info('waiting installation to finish')
        time.sleep(3)
        keyboard.send_keys("{ENTER}")
        time.sleep(20)
        keyboard.send_keys("{SPACE}")
        keyboard.send_keys("{ENTER}")
        logger.info('installation complete')
        # get the installed path
        time.sleep(1)
        installed = get_installed_path(installer_name)
        logger.info('setting bin path')

        set_apsim_bin_path(installed)
        logger.info('inserting installed path')
        insert_data(installed)
        # run tests now
        logger.info('installation complete: running test please wait....')
        run_suite(0)
    finally:
        gc.collect()


if __name__ == '__main__':
    import schedule

    schedule.every(10).seconds.do(main)

    while True:
        schedule.run_pending()
        time.sleep(1)
        # ap = main()



import gc
import shutil
import tempfile
import unittest
from pathlib import Path

from apsimNGpy.core.cs_resources import CastHelper as CastHelpers, cast_as
from apsimNGpy.core.pythonet_config import Models, is_file_format_modified
from apsimNGpy.core.model_loader import model_from_string

NEW_APSIM = is_file_format_modified()  # bool


class TestCoreModel(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        # One temp directory for the whole class
        cls.tmpdir = Path(tempfile.mkdtemp(prefix="apsim_core_cast_"))

    @classmethod
    def tearDownClass(cls):
        # Remove the whole temp tree at the end
        shutil.rmtree(cls.tmpdir, ignore_errors=True)

    def setUp(self):
        # Unique file per test to avoid cross-test interference
        self.out = self.tmpdir / f"fi_{self._testMethodName}.apsimx"
        m = model_from_string("Maize", out=str(self.out))
        self.model = getattr(m, "Model", m)

    def test_cast(self):
        """tests the cast function compiled as .dll"""
        # ensure that self.model is not the target conversion class
        if NEW_APSIM:
            self.assertNotIsInstance(self.model, Models.Core.Simulations,
                                     f'{self.model} is already a Models.Core.Simulations')
        cast = CastHelpers.CastAs[Models.Core.Simulations](self.model)
        self.assertIsInstance(cast, Models.Core.Simulations,
                              "casting to Models.Core.Simulations failed")

    def test_cast_as(self):
        """tests the cast function compiled as .dll the pythonic way"""
        if NEW_APSIM:
            # ensure that self.model is not the target conversion class
            self.assertNotIsInstance(self.model, Models.Core.Simulations,
                                     f'{self.model} is already a Models.Core.Simulations')
        cast = cast_as(model_class=Models.Core.Simulations, model=self.model)
        self.assertIsInstance(cast, Models.Core.Simulations,
                              "casting to Models.Core.Simulations failed")

    def tearDown(self):
        self.model = None
        gc.collect()

        try:
            if self.out.exists():
                self.out.unlink()
        except FileNotFoundError:
            pass


if __name__ == "__main__":
    unittest.main()



if True:

    from apsimNGpy.core.mult_cores import MultiCoreManager
    from apsimNGpy.core.apsim import ApsimModel
    from pathlib import Path

    if __name__ == "__main__":
        # create some jobs for the demo
        create_jobs = [ApsimModel('Maize').path for _ in range(1000)]  # substitute with real simulation tasks
        # initialize the multicore manager
        task_manager = MultiCoreManager(db_path=Path('test.db').resolve(), agg_func=None)
        # Run all the jobs
        task_manager.run_all_jobs(create_jobs, n_cores=16, threads=False, clear_db=True)
        # get the results
        df = task_manager.get_simulated_output(axis=0)
        # same as
        data = task_manager.results  # defaults is axis =0



import os
from apsimNGpy.core_utils.utils import timer
from apsimNGpy.core.core import Models, CoreModel, CastHelper
from apsimNGpy.core_utils.database_utils import read_db_table, get_db_table_names
from System.Collections.Generic import List
from System import Double, String
from APSIM.Core import Node
from System.IO import Path
from build.lib.apsimNGpy.core.model_tools import find_child

dotnet_list = List[String]()
dotnet_list.Add('Simulation')
path=  {}
@timer
def runner():
    # Create the APSIM model
    model = CoreModel('Maize')

    # Find simulation and datastore components
    sim = find_child(model.Simulations, 'Models.Core.Simulation', 'Simulation')
    datastore = find_child(model.Simulations, 'Models.Storage.DataStore', "DataStore")
    model.add_db_table(variable_spec=['[Maize].Total.Wt'], rename='my_table', simulation_name='Simulation', set_event_names=['[Clock].EndOfYear', '[Maize].Harvesting'])
    # Cast datastore to .NET DataStore and configure it
    cas = CastHelper.CastAs[Models.Storage.DataStore](datastore)

    cas.Open()
    cas.UseInMemoryDB = False

    # Cast simulations to IModel
    imodel = CastHelper.CastAs[Models.Core.IModel](model.Simulations)

    # Prepare the model path list (assumes dotnet_list is a List<string>)

    dotnet_list = List[String]()


    # Run the simulation using Models.Core.Run.Runner
    runner = Models.Core.Run.Runner(model.Simulations,  True, False, False, None)
    for i in dir(cas):
        print(i)
    result = runner.Run()

    # Finalize and clean up
    cas.Finalize()
    # runner.Stop()
    cs_sims = List[Models.Core.Simulation]()
    s = find_child(model.Simulations, 'Models.Core.Simulation', 'Simulation')
    cs_sims.Add(s)
    print("Progress:", runner.Progress)
    print("Status:", runner.Status)
    print("DB Path:", cas.FileName)
    print('simulations completed: ', runner.SimulationCompleted)
    print("elapsed time: ", runner.get_ElapsedTime())


    # Optionally store path in a global or passed dict

    path['p'] = cas.FileName

    # Print database contents (assumes get_db_table_names is defined elsewhere)
    print(get_db_table_names(model.datastore))


    # Close the datastore
    cas.Close()


if __name__ == '__main__':
    runner()
    model = CoreModel('Maize', out_path=os.path.abspath('xxxx.apsimx'))
    datastore = find_child(model.Simulations, 'Models.Storage.DataStore', "DataStore")
    cas =CastHelper.CastAs[Models.Storage.DataStore](datastore)


import os.path
import pathlib
import platform
import textwrap
import warnings
from pathlib import Path
from subprocess import *
from typing import Dict, Any, Union
import contextlib
import pandas as pd
from functools import lru_cache
from apsimNGpy.core.config import get_apsim_bin_path
from apsimNGpy.settings import logger
from apsimNGpy.core_utils.utils import timer
import contextlib
from apsimNGpy.settings import *
from apsimNGpy.core_utils.database_utils import read_db_table, get_db_table_names
from pathlib import Path
from typing import Union, List

apsim_bin_path = Path(get_apsim_bin_path())

# Determine executable based on OS
if platform.system() == "Windows":
    APSIM_EXEC = apsim_bin_path / "Models.exe"
else:  # Linux or macOS
    APSIM_EXEC = apsim_bin_path / "Models"


def get_apsim_version(verbose: bool = False):
    """ Display version information of the apsim model currently in the apsimNGpy config environment.

    ``verbose``: (bool) Prints the version information ``instantly``

    Example::

            apsim_version = get_apsim_version()

    """
    cmd = [APSIM_EXEC, '--version']
    if verbose:
        cmd.append("--verbose")
    res = Popen(cmd, stdout=PIPE, stderr=PIPE, text=True)
    res.wait()
    return res.stdout.readlines()[0].strip()


def upgrade_apsim_file(file: str, verbose: bool = True):
    """
    Upgrade a file to the latest version of the .apsimx file format without running the file.

    Parameters
    ---------------
    ``file``: file to be upgraded to the newest version

    ``verbose``: Write detailed messages to stdout when a conversion starts/finishes.

    ``return``
       The latest version of the .apsimx file with the same name as the input file

    Example::

        from apsimNGpy.core.base_data import load_default_simulations
        filep =load_default_simulations(simulations_object= False)# this is just an example perhaps you need to pass a lower verion file because this one is extracted from thecurrent model as the excutor
        upgrade_file =upgrade_apsim_file(filep, verbose=False)

    """
    file = str(file)
    assert os.path.isfile(file) and file.endswith(".apsimx"), f"{file} does not exists a supported apsim file"
    cmd = [APSIM_EXEC, file, '--upgrade']
    if verbose:
        cmd.append('--verbose')
    res = Popen(cmd, stdout=PIPE, stderr=PIPE, text=True)
    outp, err = res.communicate()
    if err:
        print(err)
    if verbose:
        print(outp)
    if res.returncode == 0:
        return file


def run_model_externally(model: Union[Path, str], verbose: bool = False, to_csv: bool = False) -> Popen[str]:
    """
    Runs an APSIM model externally with cross-platform support and optional CSV output.
    """
    apsim_file = model
    cmd = [str(APSIM_EXEC), str(apsim_file), '--verbose']
    if to_csv:
        cmd.append('--csv')

    with contextlib.ExitStack() as stack:
        result = stack.enter_context(Popen(cmd, stdout=PIPE, stderr=PIPE, text=True))

        try:
            out, err = result.communicate()

            if err:
                wrapped_err = textwrap.fill(err.strip(), width=80)
                logger.error("APSIM Execution Error:\n" + wrapped_err)

            if verbose:
                wrapped_out = textwrap.fill(out.strip(), width=80)
                logger.info(f"APSIM Output from {apsim_file}:\n{wrapped_out}")

            return result

        finally:
            if result.poll() is None:
                result.kill()


@lru_cache(maxsize=20)
def get_matching_files(dir_path: Union[str, Path], pattern: str, recursive: bool = False) -> List[Path]:
    """
    Search for files matching a given pattern in the specified directory.

    Args:
        ``dir_path`` (Union[str, Path]): The directory path to search in.
        ``pattern`` (str): The filename pattern to match (e.g., "*.apsimx").
        ``recursive`` (bool): If True, search recursively; otherwise, search only in the top-level directory.

    Returns:
        List[Path]: A ``list`` of matching Path objects.

    Raises:
        ``ValueError: `` If no matching files are found.
    """
    global_path = Path(dir_path)
    if recursive:
        matching_files = list(global_path.rglob(pattern))
    else:
        matching_files = list(global_path.glob(pattern))

    if not matching_files:
        raise FileNotFoundError(f"No files found that matched pattern: {pattern} in directory: {dir_path}")

    return matching_files


def collect_csv_by_model_path(model_path) -> dict[Any, Any]:
    """Collects the data from the simulated model after run
    """
    ab_path = Path(os.path.abspath(model_path))

    dirname = ab_path.parent
    name = ab_path.stem

    csv_files = list(dirname.glob(f"*{name}*.csv"))
    if csv_files:
        stems = [s.stem.split('.')[-1] for s in csv_files]
        report_paths = {k: v for k, v in zip(stems, csv_files)}

    else:
        report_paths = {}
    return report_paths


def collect_csv_from_dir(dir_path, pattern, recursive=False) -> (pd.DataFrame):
    """Collects the csf=v files in a directory using a pattern, usually the pattern resembling the one of the simulations used to generate those csv files
    ``dir_path``: (str) path where to look for csv files
    ``recursive``: (bool) whether to recursively search through the directory defaults to false:
    ``pattern``:(str) pattern of the apsim files that produced the csv files through simulations

    returns
        a generator object with pandas data frames

    Example::

         mock_data = Path.home() / 'mock_data' # this a mock directory substitute accordingly
         df1= list(collect_csv_from_dir(mock_data, '*.apsimx', recursive=True)) # collects all csf file produced by apsimx recursively
         df2= list(collect_csv_from_dir(mock_data, '*.apsimx',  recursive=False)) # collects all csf file produced by apsimx only in the specified directory directory


    """
    global_path = Path(dir_path)
    if recursive:
        matching_apsimx_patterns = global_path.rglob(pattern)
    else:
        matching_apsimx_patterns = global_path.glob(pattern)
    if matching_apsimx_patterns:
        for file_path in matching_apsimx_patterns:

            # Strip the '.apsimx' extension and prepare to match CSV files
            base_name = str(file_path.stem)

            # Search for CSV files in the same directory as the original file
            for csv_file in global_path.rglob(f"*{base_name}*.csv"):
                file = csv_file.with_suffix('.csv')
                if file.is_file():
                    df = pd.read_csv(file)
                    if isinstance(df, pd.DataFrame) and not df.empty:
                        yield df
                    else:  # notify potential issue
                        logger.warning(f"{base_name} is not a csv file or itis empty")


def run_from_dir(dir_path, pattern, verbose=False,
                 recursive=False,  # set to false because the data collector is also set to false
                 write_tocsv=True) -> [pd.DataFrame]:
    """
       This function acts as a wrapper around the ``APSIM`` command line recursive tool, automating
       the execution of APSIM simulations on all files matching a given pattern in a specified
       directory. It facilitates running simulations recursively across directories and outputs
       the results for each file are stored to a csv file in the same directory as the file'.

       What this function does is that it makes it easy to retrieve the simulated files, returning a generator that
       yields data frames

       :Parameters:
       __________________
       ``dir_path``: (str or Path, required). The path to the directory where the
           simulation files are located.
       ``pattern``: (str, required): The file pattern to match for simulation files
           (e.g., "*.apsimx").
       ``recursive``: (bool, optional):  Recursively search through subdirectories for files
           matching the file specification.
       ``write_tocsv``: (bool, optional): specify whether to write the
           simulation results to a csv. if true, the exported csv files bear the same name as the input apsimx file name
           with suffix reportname.csv. if it is ``False``,
          - if ``verbose``, the progress is printed as the elapsed time and the successfully saved csv

       ``returns``
        -- a ``generator`` that yields data frames knitted by pandas


       Example::

            mock_data = Path.home() / 'mock_data'  # As an example, let's mock some data; move the APSIM files to this directory before running
            mock_data.mkdir(parents=True, exist_ok=True)

            from apsimNGpy.core.base_data import load_default_simulations
            path_to_model = load_default_simulations(crop='maize', simulations_object=False)  # Get base model

            ap = path_to_model.replicate_file(k=10, path=mock_data) if not list(mock_data.rglob("*.apsimx")) else None

            df = run_from_dir(str(mock_data), pattern="*.apsimx", verbose=True, recursive=True)  # All files that match the pattern


       """
    dir_path = str(dir_path)

    dir_patern = f"{dir_path}/{pattern}"
    base_cmd = [str(APSIM_EXEC), str(dir_patern)]
    if recursive:
        base_cmd.append('--recursive')
    if verbose:
        base_cmd.append('--verbose')
    if write_tocsv:
        base_cmd.append('--csv')
    ran_ok = False
    with  contextlib.ExitStack() as stack:
        process = stack.enter_context(Popen(base_cmd, stdout=PIPE, stderr=PIPE,
                                            text=True))

        try:
            logger.info('waiting for APSIM simulations to complete')
            out, st_err = process.communicate()
            if verbose:
                logger.info(f'{out.strip()}')
            ran_ok = True
        finally:
            if process.poll() is None:
                process.kill()

    if write_tocsv and ran_ok:
        logger.info(f"Loading data into memory.")
        out = collect_csv_from_dir(dir_path, pattern)
        return out


@lru_cache(maxsize=None)
def apsim_executable(path, *args):
    base = [str(APSIM_EXEC), str(path), '--verbose', '--csv']
    return base


def run_p(path):
    run(apsim_executable(path))


def run(self, report_name=None,
        simulations=None,
        clean=False,
        multithread=True,
        verbose=False,
        get_dict=False, **kwargs):
    """
    Run APSIM model simulations.

    Parameters
    ----------
    report_name : str or list of str, optional
        Name(s) of report table(s) to retrieve. If not specified or missing in the database,
        the model still runs and results can be accessed later.

    simulations : list of str, optional
        Names of simulations to run. If None, all simulations are executed.

    clean : bool, default False
        If True, deletes the existing database file before running.

    multithread : bool, default True
        If True, runs simulations using multiple threads.

    verbose : bool, default False
        If True, prints diagnostic messages (e.g., missing report names).

    get_dict : bool, default False
        If True, returns results as a dictionary {table_name: DataFrame}.

    Returns
    -------
    results : DataFrame or list or dict of DataFrames
        Simulation output(s) from the specified report table(s).
    """
    try:
        # Set run type
        runtype = Models.Core.Run.Runner.RunTypeEnum.MultiThreaded if multithread \
            else Models.Core.Run.Runner.RunTypeEnum.SingleThreaded

        # Open and optionally clean the datastore
        self._DataStore.Open()
        if clean:
            self._DataStore.Dispose()
            pathlib.Path(self._DataStore.FileName).unlink(missing_ok=True)

        # Get simulations to run
        sims = self.find_simulations(simulations) if simulations else self.Simulations
        if simulations:
            cs_sims = List[Models.Core.Simulation]()
            for s in sims:
                cs_sims.Add(s)
            sim = cs_sims
        else:
            sim = sims

        # Run the model
        _run_model = Models.Core.Run.Runner(sim, True, False, False, None, runtype)
        errors = _run_model.Run()
        if errors:
            print(errors[0].ToString())

        # Determine report names
        if report_name is None:
            report_name = get_db_table_names(self.datastore)
            if '_Units' in report_name:
                report_name.remove('_Units')
            if verbose:
                warnings.warn(f"No report name specified. Using detected tables: {report_name}")

        # Read results
        if isinstance(report_name, (list, tuple)):
            if get_dict:
                results = {rep: read_db_table(self.datastore, rep) for rep in report_name}
            else:
                results = [read_db_table(self.datastore, rep) for rep in report_name]
        else:
            if get_dict:
                results = {report_name: read_db_table(self.datastore, report_name)}
            else:
                results = read_db_table(self.datastore, report_name)
    finally:
        self._DataStore.Close()

    return results



import gc
import os.path
import shutil
import time
from collections.abc import Iterable
from dataclasses import dataclass
from pathlib import Path
from typing import Union



from apsimNGpy.core_utils.utils import timer
import pandas as pd
from apsimNGpy.core.apsim import ApsimModel
from apsimNGpy.core_utils.database_utils import read_db_table, clear_all_tables, get_db_table_names, delete_all_tables, \
    delete_table
from apsimNGpy.parallel.process import custom_parallel
from sqlalchemy import MetaData, Table, Column, String, Float, Integer, MetaData
from sqlalchemy import create_engine, text

# Database connection

ID = 0


def is_my_iterable(value):
    """Check if a value is an iterable, but not a string."""
    return isinstance(value, Iterable) and not isinstance(value, str)


def simulation_exists(db_path: str, table_name: str, simulation_id: int) -> bool:
    """
    Check if a simulation_id exists in the specified table.

    Args:
        db_path (str): Path to the SQLite database file.
        table_name (str): Name of the table to query.
        simulation_id (int): ID of the simulation to check.

    Returns:
        bool: True if exists, False otherwise.
    """
    engine = create_engine(f"sqlite:///{db_path}")

    query = text(f"SELECT 1 FROM {table_name} WHERE SimulationID = :sim_id LIMIT 1")
    with engine.connect() as conn:
        result = conn.execute(query, {"sim_id": simulation_id}).first()
        return result is not None


# Example usage:
# if simulation_exists("my_database.sqlite", "results", 42):
#     print("Simulation exists!")
# else:
#     print("Simulation not found.")

setData = set()


@timer
def insert_data_with_pd(db, table, results, if_exists):
    engine = create_engine(f'sqlite:///{db}')
    results.to_sql(table, engine, index=False, if_exists=if_exists)


@dataclass(slots=True)
class MultiCoreManager:
    db_path: Union[str, Path]
    agg_func: Union[str, None] = None
    ran_ok: bool = False

    def insert_data(self, results, table):
        """
        Insert results into the specified table
        results: (Pd.DataFrame, dict) The results that will be inserted into the table
        table: str (name of the table to insert)
        """

        engine = create_engine(f"sqlite:///{str(self.db_path)}")
        metadata = MetaData()

        # Define your schema manually for full control
        if isinstance(results, pd.DataFrame):
            results_num = results.select_dtypes(include='number')

            cols = [Column(i, Float) for i in results_num.columns]
            # Find all object (string-like) columns
            str_cols_df = results.select_dtypes(include='object')

            if not str_cols_df.empty:  # safer than "is not None"
                str_cols = [Column(col_name, String) for col_name in str_cols_df.columns]
                cols.extend(str_cols)

        else:
            cols = [
                Column(col_name, String) if isinstance(results[col_name], str) else Column(col_name, Float)
                for col_name in results
            ]
        results_table = Table(
            table,
            metadata,
            Column('id', Integer, primary_key=True, autoincrement=True),
            *cols

        )
        table_meta_info = Table(
            'table_names',
            metadata,
            Column('id', Integer, primary_key=True, autoincrement=True),
            Column('table', String, primary_key=False)

        )
        table_info = {"table": table}
        # Create table if it doesn't exist
        metadata.create_all(engine)
        # if not simulation_exists(self.db_path, table, self.counter):
        # Insert data manually (append instead of replacement)
        with engine.begin() as conn:
            # Convert DataFrame rows into dicts
            if not isinstance(results, dict):
                data_dicts = results.to_dict(orient='records')
            else:
                data_dicts = results
            # keeping track of added tables

            conn.execute(results_table.insert(), data_dicts)
            conn.execute(table_meta_info.insert(), table_info)

    @property
    def tables(self):
        "Summarizes all the tables that have been created from the simulations"
        if self.ran_ok:
            if os.path.exists(self.db_path) and os.path.isfile(self.db_path) and str(self.db_path).endswith('.db'):
                dt = read_db_table(self.db_path, report_name='table_names')
                # get only unique tables
                return set(dt.table.values)
        else:
            raise ValueError("Attempting to get results from database before running all jobs")
        return None

    def run_parallel(self, model):
        """
        This is the worker for each simulation.

        The function performs two things; runs the simulation and then inserts the simulated data into a specified
        database.

        :param model: str, dict, or Path object related APSIMX json file

        returns None
        """
        # initialize the apsimNGpy model simulation engine
        _model = ApsimModel(model, out_path=None)
        table_names = _model.inspect_model('Models.Report', fullpath=False)
        # we want a unique report for each crop, because they are likely to have different report schema
        crops = _model.inspect_model('Models.PMF.Plant', fullpath=False)
        crop_table = [f"{a}_{b}" for a, b in zip(table_names, crops)]
        tables = '_'.join(crop_table)
        # run the model. without specifying report names, they will be detected automatically
        _model.run()
        # aggregate the data using the aggregated function
        if self.agg_func:
            if self.agg_func not in {'sum', 'mean', 'max', 'min', 'median', 'std'}:
                raise ValueError(f"unsupported aggregation function {self.agg_func}")
            dat = _model.results.groupby('source_table')  # if there are more than one table, we do not want to
            # aggregate them together
            out = dat.agg(self.agg_func, numeric_only=True)

            out['modelName'] = model

        else:
            out = _model.results
        # track the model id
        out['modelName'] = model
        # insert the simulated dataset into the specified database
        self.insert_data(out, table=tables)
        # clean up files related _model object
        _model.clean_up(coerce=False)
        # collect garbage before gc comes in
        gc.collect()

    def get_simulated_output(self, axis=0):
        """
        Get simulated output from the API

        :param axis: if axis =0, concatenation is along the ``rows`` and if it is 1 concatenation is along the ``columns``
        """
        if axis not in {0, 1}:
            # Errors should go silently
            raise ValueError('Wrong value for axis should be either 0 or 1')

        data = [read_db_table(self.db_path, report_name=rp) for rp in self.tables]
        return pd.concat(data, axis=axis)

    @property
    def results(self):
        """property methods for getting simulated output"""
        return self.get_simulated_output(axis=0)

    def clear_db(self):
        """Clears the database before any simulations.

          First attempt a complete ``deletion`` of the database if that fails, existing tables are all deleted"""
        if not str(self.db_path).endswith('.db'):
            raise ValueError(f"Cannot clear invalid db path: {self.db_path}")
        if os.path.exists(self.db_path):

            try:
                os.remove(self.db_path)
            except (PermissionError, FileNotFoundError):
                # if removing it has failed due to permission errors, connect to it and remove all tables
                delete_all_tables(self.db_path)

    def clear_scratch(self):
        """clears the scratch directory where apsim files are cloned before being loaded. should be called after all simulations are completed

        """
        paths = Path.cwd().glob("*scratch")
        for path in paths:
            try:
                shutil.rmtree(str(path), ignore_errors=True)
            except PermissionError:
                ...

    def run_all_jobs(self, jobs, n_cores=6, threads=False, clear_db=True, clean_up=False):
        """
        runs all provided jobs using ``processes`` or ``threads`` specified

        ``threads (bool)``: threads or processes

        ``jobs (iterable[simulations paths]``: jobs to run

        ``n_cores (int)``: number of cores to use

        ``clear_db (bool)``: clear the database existing data if any. defaults to True

        :return: None

        """
        if clear_db:
            self.clear_db()  # each simulation is fresh,
        # future updates include support for skipping some simulation
        try:
            for _ in custom_parallel(self.run_parallel, jobs, ncores=n_cores, use_threads=threads,
                                     progress_message='Processing all jobs. Please wait!: '):
                ...
            self.ran_ok = True
        finally:
            time.sleep(0.5)  # buy some time to ensure all resources are released
            if clean_up:
                self.clear_scratch()


if __name__ == '__main__':
    # quick tests
    create_jobs = (ApsimModel('Maize').path for _ in range(10))

    Parallel = MultiCoreManager(db_path='testing.db', agg_func=None)
    Parallel.run_all_jobs(create_jobs, n_cores=4, threads=False, clear_db=True, clean_up=True)
    df = Parallel.get_simulated_output(axis=0)

    insert_data_with_pd('tss.db', 'my', df, 'replace')



"""
APSIMNGpy Optimization Tutorial

This tutorial introduces both the theoretical foundation and practical application of optimization techniques using APSIMNGpy. The goal is to guide users through continuous and mixed-variable optimization problems, explaining each step as if teaching a new user or researcher in the agroecosystem modeling field.

--------------------------------------------------------------------------------

1. Overview of Optimization in Agricultural Systems

Optimization is the science of selecting the best input values (decision variables) to achieve a desired output (objective). In the context of crop modeling, this might mean finding the optimal fertilizer rate or planting density to maximize yield or minimize nutrient leaching. Optimization problems can be:

- Single-objective (e.g., maximize yield)
- Multi-objective (e.g., maximize yield while minimizing nitrate leaching)
- Continuous (variables take any value within bounds)
- Discrete or categorical (variables take on fixed options)
- Mixed (a combination of variable types)

In APSIMNGpy, we define an optimization problem, add decision variables (e.g., fertilization rate, sowing density), specify an objective function (like RMSE or mean yield), and run solvers to find optimal values.

--------------------------------------------------------------------------------

2. Tutorial Walkthrough
"""

from apsimNGpy.optimizer.single import ContinuousVariable, MixedVariable
from apsimNGpy.core.apsim import ApsimModel

# STEP 1: Load the APSIM model and observed data.
# This is typically a single simulation file you want to calibrate or optimize.
maize_model = ApsimModel("Maize")
obs = [
    7000.0, 5000.505, 1000.047, 3504.000, 7820.075,
    7000.517, 3587.101, 4000.152, 8379.435, 4000.301
]

print('Testing continuous variable optimization...')

# STEP 2A: Inheriting from ContinuousVariable
# This shows how to define a custom optimization problem class.
# You override the evaluate_objectives() method with your own objective function.
class Problem(ContinuousVariable):
    def __init__(self, apsim_model, obs):
        super().__init__(apsim_model=apsim_model)
        self.obs = obs

    def evaluate_objectives(self, **kwargs):
        # This function runs APSIM and compares the predicted results with observed data.
        predicted = self.apsim_model.run(verbose=False).results.Yield
        # Use root mean square error or another metric.
        return self.rmse(self.obs, predicted)

problem = Problem(maize_model, obs)

# STEP 2B: Alternatively, you can define the objective directly
# This is useful for simpler problems where you only need to extract something from the APSIM report table.
def maximize_yield(df):
    # Negate yield to convert to a minimization problem
    return -df.Yield.mean()

# STEP 3: Add control variables (i.e., what you want the optimizer to change).
# You can use 'add_control' to specify the path, type, and bounds.
# 'Amount' will be filled in by the optimizer. '?' marks the variable to optimize.
problem.add_control(
    path='.Simulations.Simulation.Field.Fertilise at sowing',
    Amount="?", bounds=[50, 300], v_type='qrandint', start_value=150
)

problem.add_control(
    path='.Simulations.Simulation.Field.Sow using a variable rule', CultivarName = 'B_110',
    Population="?", v_type='int', bounds=[4, 14], start_value=8
)

# # STEP 4A: Run a local optimization solver
# # This is suitable for smooth problems and quick feedback.
# res_local = problem.minimize_with_alocal_solver(
#     method='Powell',
#     options={
#         'maxiter': 100,
#         'disp': True
#     }
# )
#
# # STEP 4B: Run a global optimizer using differential evolution
# # This is useful when the surface is noisy or has many local minima.
# res_de = problem.minimize_with_de(popsize=10, maxiter=100, polish=False)
#
# print('Testing mixed variable optimization...')
#
# # STEP 5: Define a mixed-variable problem
# # MixedVariable allows combining categorical, integer, and continuous decision variables.
# problem = MixedVariable(maize_model, objectives=maximize_yield)
#
# # For a categorical variable, use 'choice' and provide a list of categories
# problem.add_control(
#     path='.Simulations.Simulation.Field.Fertilise at sowing',
#     Amount="?", v_type='choice', categories=[100, 150, 200, 250, 300], start_value=150
# )
#
# # For quantized integers, you can define a step size 'q' with 'qrandint'
# problem.add_control(
#     path='.Simulations.Simulation.Field.Sow using a variable rule',
#     Population="?", v_type='qrandint', bounds=[4, 14], start_value=8, q=2
# )
#
# # STEP 6: Run optimizers on the mixed-variable problem
# res_mixed_local = problem.minimize_with_alocal_solver(method='Powell')
# res_mixed_de = problem.minimize_with_de(popsize=20, polish=True)
#
# # STEP 7: Review results
# # Use .x, .fun, or convert to DataFrame to review the best configurations and scores.



from apsimNGpy.core.apsim import ApsimModel, Models

model = ApsimModel('Maize')
model.edit_model(
    model_type='Cultivar',
    simulations='Simulation',
    commands='[Phenology].Juvenile.Target.FixedValue',
    values=206,
    new_cultivar_name='B_110-e',
    model_name='B_110',
    cultivar_manager='Sow using a variable rule')

model.inspect_model_parameters('Cultivar', simulations='Simulation', model_name='B_110-e')
res1 = model.run().results.Yield.mean()

model.edit_model(
    model_type='Cultivar',
    simulations='Simulation',
    commands='[Phenology].Juvenile.Target.FixedValue',
    values=100,
    new_cultivar_name='B_110-e',
    model_name='B_110',
    cultivar_manager='Sow using a variable rule')

res2 = model.run().results.Yield.mean()
assert res1 != res2



from apsimNGpy.optimizer.base import AbstractProblem, VarDesc
from apsimNGpy.core.apsim import ApsimModel
from dataclasses import dataclass, field
from scipy.optimize import minimize, differential_evolution
import wrapdisc
from apsimNGpy.optimizer.optutils import edit_runner
from functools import cache
from apsimNGpy.core_utils.progbar import ProgressBar
from typing import Union
import numpy as np
from collections.abc import Iterable
SIMULATIONS = object()
from abc import abstractmethod
from apsimNGpy.optimizer.base import SING_OBJ_MIXED_VAR, SING_OBJ_CONT_VAR


new_maxiter  =0
class ContVarProblem(AbstractProblem):
    """
        Defines an optimization problem for continuous variables in APSIM simulations.

        This class enables the user to configure and solve optimization problems involving continuous
        control variables in APSIM models. It provides methods for setting up control variables,
        applying bounds and starting values, inserting variable values into APSIM model configurations,
        and running optimization routines using local solvers or differential evolution.

        Inherits from:
            AbstractProblem: A base class providing caching and model-editing functionality.

        Parameters:
            ``model (str):`` The name or path of the APSIM template file.
            .
            ``simulation (str or list, optional)``: The name(s) of the APSIM simulation(s) to target.
                                                Defaults to all simulations.

            ``decision_vars`` (list, optional): A list of VarDesc instances defining variable metadata.

            ``labels (list, optional)``: Variable labels for display and results tracking.

            ``cache_size (int):`` Maximum number of results to store in the evaluation cache.

        Attributes:
            ``model (str):`` The APSIM model template file name.

            ``simulation (str):`` Target simulation(s).

            ``decision_vars (list):`` Defined control variables.

            ``decission_vars (list):`` List of VarDesc instances for optimization.

            ``labels (list): Labels`` for variables.

            ``pbar (tqdm):`` Progress bar instance.

            ```cache (bool):`` Whether to cache evaluation results.

            ```cache_size (int):`` Size of the local cache.

        Methods:
            ``add_control(...):`` Add a new control variable to the optimization problem.

            ``bounds:`` Return the bounds for all control variables as a tuple.

            ``starting_values():`` Return the initial values for all control variables.

            ``minimize_with_local_solver(...):`` Optimize using `scipy.optimize.minimize`.

            ``optimize_with_differential_evolution(...):`` Optimize using `scipy.optimize.differential_evolution`.


        Example:
            >>> class Problem(ContVarProblem):
            ...     def evaluate(self, x):
            ...         return -self.run(verbose=False).results.Yield.mean()

            >>> problem = Problem(model="Maize", simulation="Sim")
            >>> problem.add_control("Manager", "Sow using a rule", "Population", int, 5, bounds=[2, 15])
            >>> result = problem.minimize_with_local_solver(method='Powell')
            >>> print(result.x_vars)
        """
    def __init__(self, apsim_model: 'ApsimNGpy.Core.Model', max_cache_size=400, objectives = None, decision_vars=None):

        # Initialize parent classes explicitly
        AbstractProblem.__init__(self, apsim_model, max_cache_size, objectives, decision_vars)
        self.model = apsim_model
        self.decision_vars = decision_vars or []
        self.objectives = objectives

        self.cache = True
        self.cache_size = max_cache_size
        self.pbar = None
        self.counter = 0
        self.maxiter = 0
    @property
    @abstractmethod
    def  optimization_type(self):
         return 'single'

    @property
    def bounds(self):
        return self.extract_bounds()

    def starting_values(self):
        import random
        def get_start(var):
            if 'start_value' in var:
                stat= var.get('start_value', None)
                if stat:
                     return stat
            raise ValueError(f"No start value provided for variable: {var}")

        return tuple(get_start(var) for var in self.decision_vars)
    def _insert_controls(self, x) -> None:
        x = list(x)
        for i, varR in enumerate(self.decision_vars):
            vtype = varR['v_type']
            value = x[i]
            if vtype==int or vtype =='int':
                value = int(np.rint(value))

            else:
                value = round(value, 4)
            x[i] =value
            edit_runner(self.apsim_model, decision_specs=varR, x_values=x[i])

        return x


    def _set_objective_function(self, x):
        xl = self._insert_controls(x)
        # Try local per-instance results cache first
        if self.cache and (cached := self.get_cached(*xl)):
        # Evaluation is expensive because it involves running APSIM, so the call for caching before evaluation
               return cached

        SCORE  = self.evaluate_objectives()
        if self.cache:
            self._insert_cache_result(*xl, result=SCORE)
        self._last_score = SCORE
        return SCORE


    def _open_pbar(self, labels, maxiter =400):

        #self.pbar = tqdm(total=maxiter, desc=f"Optimizing:: {', '.join(labels)}", unit=" iterations", colour="green")
        self.pbar = ProgressBar(total=maxiter, prefix=f"Optimizing:: {', '.join(labels)}", suffix='Complete', color='cyan')

    def update_pbar(self, labels, extend_by=None):
        """
        Extends the tqdm progress bar by `extend_by` steps if current progress exceeds the known max.

        Parameters:
            labels (list): List of variable labels used for tqdm description.
            extend_by (int): Number of additional steps to extend the progress bar.
        """
        total  = extend_by or int(0.4 * self.pbar.total)
        self.pbar.refresh(new_total=total)

        return self


    def _close_pbar(self):
        if self.pbar is not None:
           self.pbar.close()
    @abstractmethod
    def minimize_with_a_local_solver(self, **kwargs):
        'To be implimneted in sub class'
        pass

    @abstractmethod
    def minimize_with_de(self):
        """
        To be implimented in sub-class
        """

    def setup_obj(self):

        self.maxiter = maxiter
        self._open_pbar(labels, maxiter=maxiter)
        self.update_pbar(labels, extend_by=50)
        call_counter = {"count": 0}

        def wrapped_obj(x):
            lm = maxiter
            call_counter["count"] += 1
            xc = call_counter["count"]
            self.counter = xc
            if xc > maxiter:
                lm += 1

            self.update_pbar(labels, extend_by=50)
            self.pbar.update(1)
            # pbar.set_postfix({"score": round(self._last_score, 2)})
            return self._set_objective_function(x)
    def extract_bounds(self):
        bould_list = []
        for var in self.decision_vars:
            bounds = tuple(var.get('bounds', []))
            bould_list.append(bounds)
        return bould_list

    def evaluate_objectives(self, objectives=None):
        obj = objectives or self.objectives

        if isinstance(obj, (list, tuple)):
            if len(obj) != 1:
                raise ValueError(
                    "Only one objective function is allowed. "
                    "Combine multiple objectives into a single scalar function."
                )
            obj = obj[0]

        if not callable(obj):
            raise TypeError("Objective function must be callable.")

        self.objectives = obj
        # get results
        df = self.apsim_model.run().results
        ans = self.objectives(df)
        return ans








import numpy as np
import numpy as np
from typing import Optional, Tuple


def float_random_sampling(
        n_var: int,
        n: int,
        bounds: Optional[Tuple[np.ndarray, np.ndarray]] = None,
        **kwargs
) -> np.ndarray:
    """
    Generate random float samples in [0, 1) or within provided bounds.

    Parameters:
        n_var (int): Number of variables (dimensions).
        n (int): Number of samples.
        bounds (tuple of ndarray, optional): Tuple (xl, xu) with lower and upper bounds.
        **kwargs: Additional keyword arguments (ignored).

    Returns:
        np.ndarray: An (n, n_var) array of sampled values.
    """
    X = np.random.rand(n, n_var)

    if bounds is not None:
        xl, xu = bounds
        xl, xu = np.asarray(xl), np.asarray(xu)
        if xl.shape != (n_var,) or xu.shape != (n_var,):
            raise ValueError(f"Bounds must be arrays of shape ({n_var},), got {xl.shape} and {xu.shape}")
        if not np.all(xu >= xl):
            raise ValueError("Upper bounds must be greater than or equal to lower bounds.")
        X = xl + (xu - xl) * X

    return X


def integer_random_sampling(n_var: int, n: int, bounds: list[tuple] = None, max_default=100000):
    """
    Generate random integer samples within problem bounds.

    Parameters:
            - n_var: int, number of variables
            - bounds: (xl, xu), the lower and upper bounds as arrays e.g., ((10, 10), (100, 50)) for n-var =2
            - n: int, number of samples to generate

    Returns:
        np.ndarray: A (n_samples, n_var) array of integer samples.
    """
    if bounds is not None:
        xl, xu = bounds
        if len(xl) != len(xu):
            raise ValueError(f"lower & upper bounds must be equal")
        if len(xl) != n_var:
            raise ValueError(f"bounds does not match number of variables; expected {n_var}, got {len(xl)}")
        return np.column_stack([
            np.random.randint(xl[k], xu[k] + 1, size=n)
            for k in range(n_var)
        ])
    else:
        return np.random.randint(0, max_default, size=(n, n_var))


def boolean_random_sampling(n_var, n):
    val = np.random.random((n, n_var))
    return (val < 0.5).astype(bool)


if __name__ == '__main__':
    float_random_sampling(1, 20, [(10,), (50,)])
    boolean_random_sampling(4, 20)
    integer_random_sampling(2, 100, [(10, 10), (100, 5000)])



import json
import os
import time
import unittest
import uuid
from pathlib import Path

from apsimNGpy.core.cs_resources import CastHelper as CastHelpers
from apsimNGpy.core.model_loader import (load_apsim_model, get_model, version,
                                         save_model_to_file, load_from_path, load_crop_from_disk, load_from_dict)
from apsimNGpy.core.pythonet_config import Models
from apsimNGpy.exceptions import CastCompilationError
from apsimNGpy.tests.unittests.base_unit_tests import BaseTester


def cast(out_model, target=Models.Core.Simulations):
    casted = CastHelpers.CastAs[target](out_model)
    if casted:
        return casted
    raise CastCompilationError(" cast was not successful because the provided object is invalid")


class TestCoreModel(BaseTester):
    """Unit tests for APSIM model loading functions from various sources."""

    def setUp(self):
        self.MODEL_PATH = f"_{self._testMethodName}.apsimx"
        self.fp = Path.home() / f"{self._testMethodName}{uuid.uuid1()}_{version}.apsimx"
        self.file_on_disk = load_crop_from_disk("Soybean", out=self.MODEL_PATH)
        self.save_to = Path(f"{self._testMethodName}.apsimx")

    def mock_model_from_dict(self):
        model_to_dict = self.file_on_disk
        with open(model_to_dict) as f:
            dic = json.load(f)
            return dic

    def test_load_model_from_path_str_mtd(self):
        """Test loading APSIM model from a file path using 'string' method."""
        model_from_path = load_from_path(self.file_on_disk, method='string')
        g_model = get_model(model_from_path)
        if not isinstance(g_model, Models.Core.Simulation):
            model_class = CastHelpers.CastAs[Models.Core.Simulations](g_model)
            self.assertIsInstance(model_class, Models.Core.Simulations,
                                  'Casting failed failed loading model from file path, while using string method')
        else:
            self.skipTest(f'{g_model} is already a model loaded as {Models.Core.Simulation}')

    def test_load_model_from_path_file_mtd(self):
        """Test loading APSIM model from a file path using 'file' method."""
        model_from_path = load_from_path(self.MODEL_PATH, method='file')
        g_model = get_model(model_from_path)
        path = getattr(g_model, "FileName", "filepath.com")

        # no need for false confidence first test whether it is not the target class
        if not isinstance(g_model, Models.Core.Simulations):
            model_class = CastHelpers.CastAs[Models.Core.Simulations](g_model)
            path = getattr(model_class, "FileName", "/filepath.com")
            path = path or "/filepath.com"
            self.assertIsInstance(model_class, Models.Core.Simulations,
                                  'casting failed failed while loading model from file path using file method')
        else:
            self.skipTest(f'{g_model} is already a model loaded as {Models.Core.Simulation}')

    def test_load_model_from_dict(self):
        """Test loading APSIM model from a dictionary representation."""
        mocked_dict = self.mock_model_from_dict()
        node_or_model = load_from_dict(mocked_dict, out=self.MODEL_PATH)
        out_model = get_model(node_or_model)

        model_class = CastHelpers.CastAs[Models.Core.Simulations](out_model)
        self.assertIsInstance(model_class, Models.Core.Simulations, 'casting failed failed loading data from dict')

    def test_load_load_apsim_model_dict(self):
        """Test unified APSIM model loader with dictionary input."""
        dict_data = self.mock_model_from_dict()
        out = load_apsim_model(dict_data, out_path=self.MODEL_PATH)
        sims = cast(out.IModel)

        self.assertIsInstance(sims, Models.Core.Simulations, 'loading model fro dict using load_apsim_model '
                                                             'failed')
        self.delete_random(out.path, out.datastore)

    def test_load_load_apsim_model_path(self):
        """Test unified APSIM model loader with file path string input."""
        out = load_apsim_model(self.MODEL_PATH, out_path=self.MODEL_PATH)
        sims = cast(out.IModel)

        self.assertIsInstance(sims, Models.Core.Simulations, 'loading model from model file path using '
                                                             'load_apsim_model'
                                                             'failed')
        self.delete_random(out.path)

    @staticmethod
    def delete_random(*args):
        for file in args:
            if os.path.exists(file):
                os.remove(file)

    def test_load_load_apsim_model_path_out_is_none(self):
        """Test unified APSIM model loader with file path string input."""
        out = load_apsim_model(self.MODEL_PATH, out_path=None)
        sims = cast(out.IModel)

        self.assertIsInstance(sims, Models.Core.Simulations, 'loading model from model file path using '
                                                             'load_apsim_model and out is None'
                                                             'failed')
        # clean up from here since we don't know the random file assigned to the out path
        self.delete_random(out.path, out.datastore)

    def test_load_load_apsim_model_pathlib_path(self):
        """Test unified APSIM model loader with pathlib.Path input."""
        fp = Path(self.file_on_disk)
        out = load_apsim_model(fp, out_path=self.MODEL_PATH)
        sims = cast(out.IModel)

        self.assertIsInstance(sims, Models.Core.Simulations, 'loading model from model file pathlib path using '
                                                             'load_apsim_model'
                                                             'failed')
        self.delete_random(out.path, out.datastore)

    def test_load_load_apsim_model_path_from_disk(self):
        """Test unified APSIM model loader with pathlib.Path input."""
        fp = self.file_on_disk
        out = load_apsim_model(fp, out_path=self.MODEL_PATH)
        sims = cast(out.IModel)

        self.assertIsInstance(sims, Models.Core.Simulations, 'loading model from model file from disk using '
                                                             'load_apsim_model'
                                                             'failed')
        self.delete_random(out.path)

    def test_load_load_apsim_model_mock_none(self):
        """Test unified APSIM model loader raises ValueError when model is None."""
        fp = None
        failed = False
        try:
            load_apsim_model(fp, out_path=self.MODEL_PATH)
        except ValueError:
            failed = True

        self.assertTrue(failed, 'loading model from model while mocking None did not succeed error was not raised')

    def _test_save_method(self, mod_p):
        if os.path.exists(self.MODEL_PATH):
            os.remove(self.MODEL_PATH)
        model_to_save = load_apsim_model(mod_p, out_path=self.MODEL_PATH)
        self.MODEL_PATH = Path(self.MODEL_PATH)

        self.save_to.unlink(missing_ok=True)
        self.assertFalse(self.save_to.exists(), 'target file exists')

        save_model_to_file(model_to_save.IModel, out=self.save_to)
        exi_fp = all([self.save_to.is_file()])
        self.assertGreater(self.save_to.stat().st_size, 0, msg=f'save file failed size is zero')
        self.assertTrue(exi_fp)
        self.assertTrue(self.MODEL_PATH)
        self.delete_random(model_to_save.path)
        self.delete_random(self.MODEL_PATH)
        self.delete_random(self.save_to)

    def test_save_model_to_disk_from_default(self):

        # Test loading model from models shipped with APSIM. These reside in example dir
        # test from default
        self._test_save_method('Soybean')

    def test_save_model_to_disk_from_file(self):
        out_path = f"{self._testMethodName}.apsimx"
        # test from disk
        file_on_disk = load_crop_from_disk("Soybean", out=out_path)
        self._test_save_method(file_on_disk)
        self.delete_random(file_on_disk)

    def tearDown(self):
        if os.path.exists(self.MODEL_PATH):
            os.remove(self.MODEL_PATH)
        if os.path.exists(self.file_on_disk):
            os.remove(self.file_on_disk)


# initialize the model
if __name__ == '__main__':

    try:
        unittest.main()
    finally:
        pass



import os
import shutil
from pathlib import Path
from tempfile import TemporaryDirectory
import time
import unittest

from apsimNGpy.core.mult_cores import MultiCoreManager as ParallelRunner
from apsimNGpy.core.apsim import ApsimModel


class TestParallelRunner(unittest.TestCase):

    def _run_case(self, tmpdir: Path, agg_func, cores=2, clear=True):
        SIZE = 10
        # unique DB per case
        db_path = tmpdir / f"my_{'noagg' if agg_func is None else 'agg'}.db"

        # build jobs
        create_jobs1 = [ApsimModel('Maize').path for _ in range(SIZE)]

        # run
        runner = ParallelRunner(db_path=str(db_path), agg_func=agg_func)
        runner.run_all_jobs(create_jobs1, n_cores=cores, threads=False, clear_db=clear)

        time.sleep(0.1)  # tiny grace period for OS file handles on Windows
        df = runner.get_simulated_output(axis=0)

        shape = df.shape[0] == 10
        # quick check
        self.assertIsNotNone(df)
        if clear and agg_func is not None:
            # ensure data is deleted
            self.assertTrue(shape, "clearing the data base table failed")
        else:
            self.assertFalse(shape, "data base is clearing itself")
        # runner2.close()  # uncomment if available
        del runner

        # cleanup scratch dirs created by this run

    def test_parallel_runner_sequential(self):
        """
        Run no-aggregation first, wait for it to finish, then run aggregation.
        Ensures no overlap in processes/DB usage.
        """
        with TemporaryDirectory() as td:
            tmpdir = Path(td)

            # move CWD work into the temp dir to contain scratch/artifacts
            cwd = Path.cwd()
            try:

                # 1) no aggregation
                self._run_case(tmpdir, agg_func=None, cores=2, clear=True)
                time.sleep(1)
                # 2) aggregation (runs only after #1 completing)
                self._run_case(tmpdir, agg_func="mean", cores=2, clear=True)
            finally:

                for p in Path('.').glob("*scratch*"):
                    try:
                        shutil.rmtree(p, ignore_errors=True)
                        print('removed scratch directory', os.path.realpath(p))
                    except PermissionError:
                        pass
                time.sleep(1)


if __name__ == '__main__':
    unittest.main(verbosity=2)



import os
import tempfile
from abc import ABC, abstractmethod
from os import chdir as _chdir
from os import path
from pathlib import Path
from unittest import TestCase

import joblib

from apsimNGpy.core.base_data import load_default_simulations
from apsimNGpy.core.core import CoreModel
from apsimNGpy.settings import logger

wd = Path.cwd() / 'apsimNGpy_tests'

from apsimNGpy.settings import SCRATCH


def get_files(pattern):
    return list(Path(SCRATCH).rglob(pattern))


class BaseTester(TestCase, ABC):

    # def setUp(self):
    #     # Mock the model path and other attributes
    #     self.model_path = Path(load_default_simulations(crop='Maize', simulations_object=False), )
    #     self.model_path2 = Path(load_default_simulations(crop='Soybean', simulations_object=False), )
    #     self.logger = logger
    #     self.out_path = Path.cwd() / 'test_output.apsimx'
    #     self.test_ap_sim = CoreModel(model=self.model_path)
    #
    #     self.out = path.realpath('test_save_output.apsimx')

    # ensure all child classes implement teardown
    @abstractmethod
    def tearDown(self):
        pass

    @staticmethod
    def set_wd(_wd=None):
        if _wd is None:
            WD = Path.cwd() / 'apsimNGpy_tests'
        else:
            WD = _wd
        WD.mkdir(exist_ok=True)
        _chdir(wd)



import os
import platform
import sys as system
from apsimNGpy.core import config
from apsimNGpy.core.cs_resources import start_pythonnet
from pathlib import Path
from apsimNGpy.exceptions import ApsimBinPathConfigError
from apsimNGpy.core_utils.utils import timer

APSIM_BIN_PATH = config.get_apsim_bin_path()

start_pythonnet()


def is_file_format_modified():
    """
    Checks if the APSIM.CORE.dll is present in the bin path
    @return: bool
    """
    patterns = {"*APSIM.CORE.dll", "*APSIM.Core.dll"}
    path = []
    for pat in patterns:
        path.extend(Path(APSIM_BIN_PATH).rglob(pat))
    if len(path) > 0:
        return True
    return False


def load_pythonnet(bin_path=APSIM_BIN_PATH):
    """
    A method for loading Python for .NET (pythonnet) and APSIM models.

    This class provides a callable method for initializing the Python for .NET (pythonnet) runtime and loading APSIM models.
    Initialize the Python for .NET (pythonnet) runtime and load APSIM models.

        This method attempts to load the 'coreclr' runtime, and if not found, falls back to an alternate runtime.
        It also sets the APSIM binary path, adds necessary references, and returns a reference to the loaded APSIM models.

        Returns:
        -------
        lm: Reference to the loaded APSIM models

        Raises:
        ------
        KeyError: If APSIM path is not found in the system environmental variable.
        ValueError: If the provided APSIM path is invalid.

        Notes:
        It raises a KeyError if APSIM path is not found. Please edit the system environmental variable on your computer.
    Attributes:
    ----------
    None
    """

    if not os.path.isdir(bin_path):
        raise ApsimBinPathConfigError(
            "Bin path configuration error or APSIM is not yet installed, or APSIM was recently uninstalled")

    candidate = os.path.join(bin_path, 'bin')
    if os.path.basename(bin_path).lower() != 'bin' and os.path.isdir(candidate):
        bin_path = candidate

    system.path.append(bin_path)

    import clr
    start_pythonnet()
    clr.AddReference("System")
    # model_path = os.path.join(bin_path, 'Models.dll')
    # clr.AddReference(model_path)
    # apsimNG = clr.AddReference('ApsimNG')
    clr.AddReference("Models")
    if is_file_format_modified():
        clr.AddReference('APSIM.Core')

    # return lm, sys, pythonnet.get_runtime_info()


load_pythonnet()
# now we can safely import C# libraries

from System.Collections.Generic import *
from Models.Core import Simulation

import Models

from System import *

Models = Models



def get_apsim_file_reader(method: str = 'string'):
    if is_file_format_modified() or not getattr(Models.Core.ApsimFile, "FileFormat", None):
        import APSIM.Core
        base = APSIM.Core.FileFormat

    else:
        base = Models.Core.ApsimFile.FileFormat
    match method:
        case 'string':
            return getattr(base, 'ReadFromString')
        case 'file':
            return getattr(base, 'ReadFromFile')
        case _:
            raise NotImplementedError(f"{method} method is not implemented")


def get_apsim_file_writer():
    if is_file_format_modified() or not getattr(Models.Core.ApsimFile, "FileFormat", None):
        import APSIM.Core
        base = APSIM.Core.FileFormat
        os.environ['A'] = 'true'
    else:
        base = Models.Core.ApsimFile.FileFormat
    return getattr(base, 'WriteToString')


# Example usage:
if __name__ == '__main__':
    loader = load_pythonnet()
    loaded_models = loader



"""This is on trial as a module to replace some fucntionalities provided by using pythonnet API"""
import gc
import json
import os
import sys
from apsimNGpy.core.model_loader import load_as_dict
from apsimNGpy.core.config import load_crop_from_disk
from os.path import realpath
from collections import OrderedDict
from apsimNGpy.core.model_tools import validate_model_obj, Models

path2file = load_crop_from_disk('Maize', '../t.apsim')
f_name = realpath(path2file)
import datetime
import re

temp= r"C:\Users\vanguard\Downloads\maize.apsimx"
with open(temp, "r+", encoding='utf-8') as apsimx:
    data = OrderedDict()
    app_ap = data | json.load(apsimx)


def match_model_path(model_string):
    pattern = r'\b(?:[A-Za-z_][A-Za-z0-9_]*\.){1,}[A-Za-z_][A-Za-z0-9_]*\b'
    out = re.findall(pattern, model_string)
    if out:
        return out[0]


def evaluate_model_class(model_class):
    """
    make sure the provided model_class is an attributes of Models namespace.

    facilitates users in supplying only the name of the model class e.g., Clock instead of Models.Clock

    @param model_class:
    @return: full string representation of model_class
    """
    evaluated = validate_model_obj(model_class)
    return match_model_path(str(evaluated))


class JsonUtilities:
    def __init__(self, _mod=f_name):
        self.json_name = _mod
        self.json = self.load_json(f_name)

    @staticmethod
    def load_json(jpath):

        with open(jpath, "r+", encoding='utf-8') as json_load:
            order = OrderedDict()
            merge = order | json.load(json_load)
        return merge

    def get_children(self, scope=None, recursive=False):
        """
        Recursively extract all 'Children' nodes from the JSON tree.
        """
        scope = scope or self.json
        children = scope.get('Children', [])
        if not recursive:
            return children
        all_children = []

        for child in children:
            all_children.append(child)
            # Recursively collect from nested children
            all_children.extend(self.get_children(child))

        return all_children

    def get_simulations(self):
        """
        Recursively collect all objects of type 'Models.Core.Simulation'.
        Returns a dictionary keyed by model path.
        """
        all_nodes = self.get_children(recursive=True)
        sims = {}

        for node in all_nodes:
            model_type = node.get('$type', '')
            matches = match_model_path('Models.Core.Simulation')
            if matches and matches == 'Models.Core.Simulation':
                sims[matches] = node

        return sims

    def find_all_descendants(self, model_class, node=None):
        """
        Recursively collect all in the specified ``model_class`` objects.
        Returns a dictionary with paths as keys and node objects as values.
        """
        _model_class = evaluate_model_class(model_class)
        if not match_model_path(_model_class):
            raise ValueError('un able to proceed')
        if node is None:
            node = self.json

        def _recurse(node, path=''):
            sims = {}
            node_name = node.get('Name', '<unnamed>')
            current_path = f"{path}.{node_name}" if path else f'.{node_name}'

            model_type = node.get('$type', '')
            matches = match_model_path(model_type)

            if matches and matches == _model_class:
                sims[current_path] = node

            for child in self.get_children(scope=node, recursive=True):
                sims.update(_recurse(child, current_path))

            return sims

        return _recurse(node)

    def get_fullpath(self, model_class):
        model_class = evaluate_model_class(model_class)
        data = self.find_all_descendants(model_class)
        return tuple(data.keys())

    def edit_model_by_path(self, model_class, path, verbose=True, **kwargs):
        model = self.find_all_descendants(model_class=model_class).get(path)
        match model_class:
            case "Models.Manager":
                parameters = model.get('Parameters', {})
                kas = set(kwargs.keys())
                kav = {i['Key'] for i in parameters}
                dif = kas - kav
                if len(dif) > 0:
                    raise ValueError(f"{kas - kav} is not a valid parameter for {path}")
                # Manager scripts have extra attribute parameters

                for i in parameters:
                    param = i['Key']
                    if param in kas:
                        param_value = kwargs[param]
                        i['Value'] = param_value
                return model

            case "Models.Clock" | Models.Clock:
                validated = dict(End='End', Start='Start', end='End', start='Start', end_date='End',
                                 start_date='Start')

                for kwa, value in kwargs.items():
                    key = validated.get(kwa, 'unknown')  # APSIM uses camelcase
                    if key in ['End', 'Start']:
                        parsed_value = datetime.datetime.fromisoformat(value)
                        model[key] = parsed_value.isoformat()
                        if verbose:
                            print(f"Set {key} to {parsed_value}")

                    else:
                        raise AttributeError(
                            f"no valid Clock attributes were passed. Valid arguments are: '{", ".join(validated.keys())}'")

            case "Models.Climate.Weather" | Models.Climate.Weather:
                met_file = kwargs.get('weather_file') or kwargs.get('met_file')
                if met_file is None:
                    raise ValueError('Use key word argument "weather_file" or "met_file" to supply the weather data')
                    # To avoid carrying over a silent bug or waiting for the bug to manifest during a model run,
                    # there is a need to raise here
                if not os.path.exists(met_file):
                    raise FileNotFoundError(f"'{met_file}' rejected because it does not exist on the computer")
                if not os.path.isfile(met_file):
                    raise FileNotFoundError(f"'{met_file}' is not a valid file did you forget to add .met at the end?")
                c_file = model.get('FileName', '')
                model['FileName'] = met_file
                if verbose:
                    print(f"weather file changed from '{c_file}' to '{met_file}'")

        return model
    def read(self):
        json_string = json.dumps(self.json)

        model_frm_str = Models.Core.ApsimFile.FileFormat.ReadFromString[Models.Core.Simulations](json_string, None,
                                                                                         True,
                                                                                         fileName='./_.apsimx')
        return model_frm_str


if __name__ == '__main__':
    sys.setrecursionlimit(9000)
    jj = JsonUtilities(app_ap)
    ch = jj.get_children()
    jj.find_all_descendants('Clock')
    jj.get_fullpath('Clock')
    jj.get_fullpath('Models.PMF.Cultivar')
    print(gc.collect())



import unittest
from apsimNGpy.core.apsim import ApsimModel
from apsimNGpy.tests.unittests.base_unit_tests import BaseTester
import os


class TestWeatherDownload(BaseTester):
    """Test weather data download functionality"""

    # Set up the model to use
    def setUp(self):
        self.model = ApsimModel('Maize')
        self.out = 'test_edit_model.apsimx'
        self.model.edit_model(
            model_type='Clock',
            model_name='Clock',
            simulations='Simulation',
            start='1990-01-01',
            end='2000-01-01'
        )
        self.LONLAT = 35.582520, 0.347596
        self.test_filename = 'test_nasa_download.met'
        self.default_filename = 'Maize_nasa_2000_2001.met'
        # Clean up any existing test files
        if os.path.exists(self.test_filename):
            os.remove(self.test_filename)

    def tearDown(self):
        """Cleanup work after testing"""
        # Clean up test files
        if os.path.exists(self.test_filename):
            os.remove(self.test_filename)
        if os.path.exists(self.default_filename):
            os.remove(self.default_filename)

    def test_weather_file_download_with_specified_filename(self):
        """Test if downloaded met file is saved with user-specified filename"""
        # Ensure test file doesn't exist
        self.assertFalse(os.path.exists(self.test_filename),
                         f"Test file {self.test_filename} should not exist before testing")

        # Download with specified filename
        self.model.get_weather_from_web(
            lonlat=self.LONLAT,
            start=2000,
            end=2001,
            simulations='Simulation',
            source='nasa',
            filename=self.test_filename
        )

        # Check if specified filename is used
        inspect_met_file = self.model.inspect_model_parameters(model_type='Weather', model_name='Weather',
                                                               simulations='Simulation')
        self.assertEqual(self.test_filename, os.path.basename(inspect_met_file),
                         f"Inspected met file should be {self.test_filename}")

        # Verify file is created with specified filename
        self.assertTrue(os.path.exists(self.test_filename),
                        f"Downloaded met file should be saved with specified filename {self.test_filename}")

        # Verify file is not empty
        self.assertGreater(os.path.getsize(self.test_filename), 0,
                           "Downloaded met file should not be empty")

        # Don't check file content, only verify file exists and is not empty

    def test_weather_file_download_default_filename(self):
        """Test default behavior when filename is not specified"""
        # Ensure test file doesn't exist
        self.assertFalse(os.path.exists(self.default_filename),
                         f"Test file {self.default_filename} should not exist before testing")

        # Download with default filename
        self.model.get_weather_from_web(
            lonlat=self.LONLAT,
            start=2000,
            end=2001,
            simulations='Simulation',
            source='nasa',

        )

        # Check if the default filename is used
        inspect_met_file = self.model.inspect_model_parameters(model_type='Weather', model_name='Weather',
                                                               simulations='Simulation')
        self.assertEqual(self.default_filename, os.path.basename(inspect_met_file),
                         f"Inspected met file should be {self.default_filename}")

        # Check if default met file is created (usually has default naming rules)
        self.assertTrue(os.path.exists(self.default_filename),
                        f"Downloaded met file should be saved with default filename {self.default_filename}")

        # There should be at least one met file
        self.assertGreater(os.path.getsize(self.default_filename), 0, "Downloaded met file should not be empty")


if __name__ == '__main__':
    unittest.main()



from apsimNGpy.optimizer._one_obj import ContVarProblem, SING_OBJ_CONT_VAR, SING_OBJ_MIXED_VAR
from apsimNGpy.optimizer._mixed import MixedVarProblem
class ContinuousVariable(ContVarProblem):
    def __init__(self, apsim_model: 'ApsimNGpy.Core.Model', max_cache_size=400, objectives=None, decision_vars=None):
        # Initialize parent classes explicitly
        ContVarProblem.__init__(self, apsim_model, max_cache_size, objectives, decision_vars)
        self.model = apsim_model
        self.decision_vars = decision_vars or []
        self.objectives = objectives

        self.cache = True
        self.cache_size = max_cache_size
        self.pbar = None
        self.counter = 0
        self.maxiter = 0

    def optimimization_type(self):
        return SING_OBJ_CONT_VAR

    def minimize_with_alocal_solver(self, **kwargs):
        """
        Run a local optimization solver using `scipy.optimize.minimize`.

        This method wraps ``scipy.optimize.minimize`` to solve APSIM optimization problems
        defined using APSIM control variables and variable encodings. It tracks optimization progress via a progress bar,
        and decodes results into user-friendly labeled dictionaries.

        Optimization methods available in `scipy.optimize.minimize` include:

        +------------------+------------------------+-------------------+----------------+---------------------+----------------------------------------------+
        | Method           | Type                   | Gradient Required | Handles Bounds | Handles Constraints | Notes                                        |
        +==================+========================+===================+================+=====================+==============================================+
        | Nelder-Mead      | Local (Derivative-free)| No                | No             | No                  | Simplex algorithm                            |
        +------------------+------------------------+-------------------+----------------+---------------------+----------------------------------------------+
        | Powell           | Local (Derivative-free)| No                | Yes            | No                  | Direction set method                         |
        +------------------+------------------------+-------------------+----------------+---------------------+----------------------------------------------+
        | CG               | Local (Gradient-based) | Yes               | No             | No                  | Conjugate Gradient                           |
        +------------------+------------------------+-------------------+----------------+---------------------+----------------------------------------------+
        | BFGS             | Local (Gradient-based) | Yes               | No             | No                  | Quasi-Newton                                 |
        +------------------+------------------------+-------------------+----------------+---------------------+----------------------------------------------+
        | Newton-CG        | Local (Gradient-based) | Yes               | No             | No                  | Newton's method                              |
        +------------------+------------------------+-------------------+----------------+---------------------+----------------------------------------------+
        | L-BFGS-B         | Local (Gradient-based) | Yes               | Yes            | No                  | Limited memory BFGS                          |
        +------------------+------------------------+-------------------+----------------+---------------------+----------------------------------------------+
        | TNC              | Local (Gradient-based) | Yes               | Yes            | No                  | Truncated Newton                             |
        +------------------+------------------------+-------------------+----------------+---------------------+----------------------------------------------+
        | COBYLA           | Local (Derivative-free)| No                | No             | Yes                 | Constrained optimization by linear approx.   |
        +------------------+------------------------+-------------------+----------------+---------------------+----------------------------------------------+
        | SLSQP            | Local (Gradient-based) | Yes               | Yes            | Yes                 | Sequential Least Squares Programming         |
        +------------------+------------------------+-------------------+----------------+---------------------+----------------------------------------------+
        | trust-constr     | Local (Gradient-based) | Yes               | Yes            | Yes                 | Trust-region constrained                     |
        +------------------+------------------------+-------------------+----------------+---------------------+----------------------------------------------+
        | dogleg           | Local (Gradient-based) | Yes               | No             | No                  | Requires Hessian                             |
        +------------------+------------------------+-------------------+----------------+---------------------+----------------------------------------------+
        | trust-ncg        | Local (Gradient-based) | Yes               | No             | No                  | Newton-CG trust region                       |
        +------------------+------------------------+-------------------+----------------+---------------------+----------------------------------------------+
        | trust-exact      | Local (Gradient-based) | Yes               | No             | No                  | Trust-region, exact Hessian                  |
        +------------------+------------------------+-------------------+----------------+---------------------+----------------------------------------------+
        | trust-krylov     | Local (Gradient-based) | Yes               | No             | No                  | Trust-region, Hessian-free                   |
        +------------------+------------------------+-------------------+----------------+---------------------+----------------------------------------------+

        Reference:

        https://docs.scipy.org/doc/scipy/reference/generated/scipy.optimize.minimize.html#scipy.optimize.minimize.

        Parameters::

        **kwargs:

            Arbitrary keyword arguments passed to `scipy.optimize.minimize`, such as:

            - ``method (str)``: The optimization method to use.

            - ``options (dict)``: Solver-specific options like `disp`, `maxiter`, `gtol`, etc.

            - ``bounds (list of tuple)``: Variable bounds; defaults to self.bounds if not provided.

            - ``x0 (list):`` Optional starting guess (will override default provided values with ``add_control_var`` starting values).

        Returns:
            result (OptimizeResult):
                The optimization result object with the following additional field:
                - result.x_vars (dict): A dictionary of variable labels and optimized values.

        Example::

          from apsimNGpy.optimizer.one_objective import ContVarProblem
          class Problem(ContVarProblem):
                def __init__(self, model=None, simulation='Simulation'):
                    super().__init__(model, simulation)
                    self.simulation = simulation
                def evaluate(self, x, **kwargs):
                   return -self.run(verbose=False).results.Yield.mean()

          problem = Problem(model="Maize", simulation="Sim")
          problem.add_control("Manager", "Sow using a rule", "Population", vtype="grid",
                                start_value=5, values=[5, 9, 11])
          problem.add_control("Manager", "Sow using a rule", "RowSpacing", vtype="grid",
                                start_value=400, values=[400, 800, 1200])
          result = problem.minimize_with_local_solver(method='Powell', options={"maxiter": 300})
          print(result.x_vars)
          {'Population': 9, 'RowSpacing': 800}
        """

        try:
            x0 = kwargs.pop("x0", [1] * len(self.decision_vars))
            if 'bounds' not in kwargs.keys():
                kwargs['bounds'] = self.bounds
            max_iter = kwargs.get("options", {}).get("maxiter", 400)

            self._open_pbar(self.labels, maxiter=max_iter)
            call_counter = {"count": 0}
            def wrapped_obj(x):
                call_counter["count"] += 1
                self.pbar.update(1)
                #pbar.set_postfix({"score": round(self._last_score, 2)})
                return self._set_objective_function(x)

            result = minimize(wrapped_obj, x0=self.starting_values(), **kwargs)
            result.x_vars = dict(zip(self.labels, result.x))
            self.outcomes = result
            return result
        finally:
            self._close_pbar()
            self.clear_cache()

    def optimization_type(self):
        return SING_OBJ_CONT_VAR
    def minimize_with_de(self, args=(), strategy='best1bin',
                                             maxiter=1000, popsize=15,
                                             tol=0.01, mutation=(0.5, 1), recombination=0.7,
                                             rng=None, callback=None,
                                             disp=True, polish=True,
                                             init='latinhypercube', atol=0,
                                             updating='immediate', workers=1,
                                             constraints=(), x0=None, *,
                                             integrality=None, vectorized=False):
        """

        reference; https://docs.scipy.org/doc/scipy/reference/generated/scipy.optimize.differential_evolution.html
        """

        try:

            self.maxiter = maxiter
            self._open_pbar(self.labels, maxiter=maxiter)
            self.update_pbar(self.labels, extend_by=50)
            call_counter = {"count": 0}


            def wrapped_obj(x):
                lm =maxiter
                call_counter["count"] += 1
                xc  =  call_counter["count"]
                self.counter = xc
                if xc > maxiter:
                    lm +=1

                self.update_pbar(self.labels, extend_by=50)
                self.pbar.update(1)
                #pbar.set_postfix({"score": round(self._last_score, 2)})
                return self._set_objective_function(x)

            result = differential_evolution(wrapped_obj, self.bounds, args=(), strategy='best1bin',
                                      maxiter=maxiter, popsize=popsize, tol=tol,
                                      mutation=mutation, recombination=recombination,
                                       callback=callback, disp=disp,
                                      polish=polish, init=init,
                                      atol=atol, updating=updating,
                                      workers=workers, constraints=constraints)

            result.x_vars = dict(zip(self.labels, result.x))
            self.outcomes =result
            return result
        finally:
            self._close_pbar()
            self.clear_cache()

class MixedVariable(MixedVarProblem):
    def __init__(self, apsim_model: 'ApsimNGpy.Core.Model', max_cache_size=400, objectives=None, decision_vars=None):
        # Initialize parent classes explicitly
        super.__init__(self, apsim_model=apsim_model, max_cache_size=max_cache_size, objectives =objectives, decision_vars=decision_vars)
        self.model = apsim_model
        self.decision_vars = decision_vars or []
        self.objectives = objectives

        self.cache = True
        self.cache_size = max_cache_size
        self.pbar = None
        self.counter = 0
        self.maxiter = 0
    def minimize_with_alocal_solver(self, **kwargs):
        """
            Run a local optimization solver (e.g., Powell, L-BFGS-B, etc.) on given defined problem.

            This method wraps ``scipy.optimize.minimize`` and handles mixed-variable encoding internally
            using the `Objective` wrapper from ``wrapdisc``. It supports any method supported by SciPy's
            `minimize` function and uses the encoded starting values and variable bounds. This decoding implies that you can optimize categorical variable such as start dates or
            cultivar paramter with xy numerical values.

            Progress is tracked using a progress bar, and results are automatically decoded and stored
            in ``self.outcomes``.

            Parameters:
                **kwargs: Keyword arguments passed directly to `scipy.optimize.minimize`.
                          Important keys include:
                            - ``method (str)``: Optimization algorithm (e.g., 'Powell', 'L-BFGS-B').
                            - ``options (dict)``: Dictionary of solver options like maxiter, disp, etc.
        scipy.optimize.minimize provide a number of optimization algorithms see table below or for details check their website:
        https://docs.scipy.org/doc/scipy/reference/generated/scipy.optimize.minimize.html#scipy.optimize.minimize

        +------------------+------------------------+-------------------+----------------+---------------------+----------------------------------------------+
        | Method           | Type                   | Gradient Required | Handles Bounds | Handles Constraints | Notes                                        |
        +==================+========================+===================+================+=====================+==============================================+
        | Nelder-Mead      | Local (Derivative-free)| No                | No             | No                  | Simplex algorithm                            |
        +------------------+------------------------+-------------------+----------------+---------------------+----------------------------------------------+
        | Powell           | Local (Derivative-free)| No                | Yes            | No                  | Direction set method                         |
        +------------------+------------------------+-------------------+----------------+---------------------+----------------------------------------------+
        | CG               | Local (Gradient-based) | Yes               | No             | No                  | Conjugate Gradient                           |
        +------------------+------------------------+-------------------+----------------+---------------------+----------------------------------------------+
        | BFGS             | Local (Gradient-based) | Yes               | No             | No                  | Quasi-Newton                                 |
        +------------------+------------------------+-------------------+----------------+---------------------+----------------------------------------------+
        | Newton-CG        | Local (Gradient-based) | Yes               | No             | No                  | Newton's method                              |
        +------------------+------------------------+-------------------+----------------+---------------------+----------------------------------------------+
        | L-BFGS-B         | Local (Gradient-based) | Yes               | Yes            | No                  | Limited memory BFGS                          |
        +------------------+------------------------+-------------------+----------------+---------------------+----------------------------------------------+
        | TNC              | Local (Gradient-based) | Yes               | Yes            | No                  | Truncated Newton                             |
        +------------------+------------------------+-------------------+----------------+---------------------+----------------------------------------------+
        | COBYLA           | Local (Derivative-free)| No                | No             | Yes                 | Constrained optimization by linear approx.   |
        +------------------+------------------------+-------------------+----------------+---------------------+----------------------------------------------+
        | SLSQP            | Local (Gradient-based) | Yes               | Yes            | Yes                 | Sequential Least Squares Programming         |
        +------------------+------------------------+-------------------+----------------+---------------------+----------------------------------------------+
        | trust-constr     | Local (Gradient-based) | Yes               | Yes            | Yes                 | Trust-region constrained                     |
        +------------------+------------------------+-------------------+----------------+---------------------+----------------------------------------------+
        | dogleg           | Local (Gradient-based) | Yes               | No             | No                  | Requires Hessian                             |
        +------------------+------------------------+-------------------+----------------+---------------------+----------------------------------------------+
        | trust-ncg        | Local (Gradient-based) | Yes               | No             | No                  | Newton-CG trust region                       |
        +------------------+------------------------+-------------------+----------------+---------------------+----------------------------------------------+
        | trust-exact      | Local (Gradient-based) | Yes               | No             | No                  | Trust-region, exact Hessian                  |
        +------------------+------------------------+-------------------+----------------+---------------------+----------------------------------------------+
        | trust-krylov     | Local (Gradient-based) | Yes               | No             | No                  | Trust-region, Hessian-free                   |
        +------------------+------------------------+-------------------+----------------+---------------------+----------------------------------------------+

            Returns:
                result (OptimizeResult): The result of the optimization, with an additional
                                         `x_vars` attribute that provides a labeled dict of optimized
                                         control variable values.

            Raises:
                Any exceptions raised by `scipy.optimize.minimize`.

            Example:
            --------
            The following example shows how to use this method, the evaluation is very basic, but you
            can add a more advanced evaluation by adding a loss function e.g RMSE os NSE by comparing with the observed and predicted,
            and changing the control variables::

            class Problem(MixedVarProblem):
                def __init__(self, model=None, simulation='Simulation'):
                    super().__init__(model, simulation)
                    self.simulation = simulation

                def evaluate(self, x, **kwargs):
                    # All evlauations can be defined inside here, by taking into accound the fact that the results object returns a data frame
                    # Also, you can specify the database table or report name holding the ``results``
                    return -self.run(verbose=False).results.Yield.mean() # A return is based on your objective definition, but as I said this could a ``RRMSE`` error or any other loss function

            # Ready to initialise the problem

            .. code-block:: python

                 problem.add_control(
                    path='.Simulations.Simulation.Field.Fertilise at sowing',
                    Amount="?",
                    bounds=[50, 300],
                    v_type="float",
                    start_value =50
                 )

                problem.add_control(
                    path='.Simulations.Simulation.Field.Sow using a variable rule',
                    Population="?",
                    bounds=[4, 14],
                    v_type="float",
                    start_value=5
                )

            """
        try:

            self._open_pbar(labels= self.labels)
            call_counter = {"count": 0}

            def wrapped_obj(x, *args):
                call_counter["count"] += 1
                self.pbar.update(1)
                return self._set_objective_function(x)

            wrapped = Objective(
                func=wrapped_obj,
                variables=self._variables
            )

            bounds = wrapped.bounds

            # Handle initial guess
            initial_guess = self.starting_values()

            encoded_initial = wrapped.encode(initial_guess)

            result = minimize(wrapped, x0=encoded_initial, bounds=bounds, **kwargs)

            # Attach labeled solution
            decoded_solution = wrapped.decode(result.x)
            result.x_vars = dict(zip(self.labels, decoded_solution))
            self.outcomes = result
            return result

        finally:
            self.clear_cache()
            self._close_pbar()

    def minimize_with_de(self,
                                             args=(),
                                             strategy='best1bin',
                                             maxiter=1000,
                                             popsize=15,
                                             tol=0.01,
                                             mutation=(0.5, 1),
                                             recombination=0.7,
                                             rng=None,
                                             callback=None,
                                             disp=True,
                                             polish=True,
                                             init='latinhypercube',
                                             atol=0,
                                             updating='immediate',
                                             workers=1,
                                             constraints=(),
                                             x0=None,
                                             seed =1,
                                             *,
                                             integrality=None,
                                             vectorized=False):
        """
        Runs differential evolution on the wrapped objective function.
        Reference: https://docs.scipy.org/doc/scipy/reference/generated/scipy.optimize.differential_evolution.html
        """

        try:

            self._open_pbar(labels=self.labels)
            call_counter = {"count": 0}
            func = getattr(self, '_set_objective_function')

            def wrapped_obj(x, *args):
                call_counter["count"] += 1
                self.pbar.update(1)
                return self._set_objective_function(x)

            wrapped = Objective(
                func=wrapped_obj,
                variables=self._variables
            )

            bounds = wrapped.bounds

            # Handle initial guess
            initial_guess = self.starting_values()

            encoded_initial = wrapped.encode(initial_guess)

            # Run optimization
            result = differential_evolution(
                wrapped,
                bounds=wrapped.bounds,
                args=args,
                strategy=strategy,
                maxiter=maxiter,
                popsize=popsize,
                tol=tol,
                mutation=mutation,
                recombination=recombination,
                seed=seed,
                disp=disp,
                polish=polish,
                init=init,
                atol=atol,
                updating=updating,
                workers=workers,
                constraints=constraints,
                x0=encoded_initial,
                integrality=integrality,
                vectorized=vectorized
            )

            # Attach a labeled solution
            decoded_solution = wrapped.decode(result.x)
            result.x_vars = dict(zip(self.labels, decoded_solution))
            self.outcomes = result
            return result

        finally:
            self.clear_cache()
            self._close_pbar()

    def optimization_type(self):
        return SING_OBJ_MIXED_VAR

    # tests
if __name__ == '__main__':
    from apsimNGpy.core.apsim import ApsimModel
    maize_model = "Maize"
    obs = [
        7000.0,
        5000.505,
        1000.047,
        3504.000,
        7820.075,
        7000.517,
        3587.101,
        4000.152,
        8379.435,
        4000.301
    ]


    class Problem(ContinuousVariable):
        def __init__(self, apsim_model: 'ApsimNGpy.Core.Model', obs, max_cache_size=400, objectives=None,
                     decision_vars=None):
            super().__init__(apsim_model=apsim_model, max_cache_size=max_cache_size, objectives=objectives,
                             decision_vars=decision_vars)
            self.obs = obs

        def evaluate_objectives(self, **kwargs):
            # set up everything you need here
            predicted = self.apsim_model.run(verbose=False).results.Yield

            ans = self.rmse(self.obs, predicted)

            return -predicted.mean()


    def maximize_yield(df):
        ans = -df.Yield.mean()
        print(ans)
        return ans



    maize_model = ApsimModel('Maize')
    problem = Problem(maize_model, obs, objectives=maximize_yield)
    problem.add_control(
        **{'path': '.Simulations.Simulation.Field.Fertilise at sowing', 'Amount': "?", "bounds": [50, 300],
           "v_type": "qrandint"}, start_value=150)
    problem.add_control(
        **{'path': '.Simulations.Simulation.Field.Sow using a variable rule', 'Population': "?", 'v_type': 'int',
           'bounds': [4, 14]}, start_value=8)
    # problem.add_control('Manager', "Sow using a variable rule", 'Population',  int,5, bounds=[2, 15])
    # problem.add_control('Manager', "Sow using a variable rule", 'RowSpacing', int, 500, bounds=[400, 800])

    # res = problem.minimize_with_local_solver( method  ='Powell',  options={
    #     # 'xatol': 1e-4,      # absolute error in xopt between iterations
    #     # 'fatol': 1e-4,      # absolute error in func(xopt) between iterations
    #     'maxiter': 100,    # maximum number of iterations
    #     'disp': True ,      # display optimization messages
    #
    # })
    # de = problem.optimize_with_differential_evolution(popsize=10, maxiter=100)
    dep = problem.minimize_with_differential_evolution(popsize=10, maxiter=100, polish=False)



import os
import unittest
from pathlib import Path

from apsimNGpy.core.apsim import ApsimModel
from apsimNGpy.tests.unittests.base_unit_tests import BaseTester
from apsimNGpy.core_utils.clean import clean
import tempfile

wd = Path.cwd() / "test_apsim"
wd.mkdir(parents=True, exist_ok=True)
os.chdir(wd)


class TestCoreModel(BaseTester):
    def setUp(self):
        self.out_path = Path(f"{self._testMethodName}.apsimx")
        self.test_ap_sim = ApsimModel("Maize", out_path=self.out_path)

    def test_run(self):
        self.test_ap_sim.run()
        self.assertTrue(self.test_ap_sim.ran_ok)
        self.test_ap_sim.clean_up(db=True)

    def test_clone_model(self):
        from apsimNGpy.core.config import BASE_RELEASE_NO, APSIM_VERSION_NO
        model = self.test_ap_sim
        sim_name = 'clone_test'
        if APSIM_VERSION_NO > BASE_RELEASE_NO:
            self.skipTest(f"This version of apsimNGpy {APSIM_VERSION_NO} is not supported by clone_method becasue"
                          f"it is greater than version {BASE_RELEASE_NO}")
        model.clone_model('Models.Core.Simulation', 'Simulation',
                          'Models.Core.Simulations', rename=sim_name)
        model.run()
        self.assertTrue(model.ran_ok)
        # inspect simulations
        sims = model.inspect_model(model_type='Models.Core.Simulation', fullpath=False)
        assert sim_name in sims, (f'{sim_name} is not among the current simulations, implying simulation was not '
                                  f'successful')
        model.clean_up(db=True)

    def tearDown(self):
        pass


# initialize the model
if __name__ == '__main__':
    unittest.main()



import os

import requests
from bs4 import BeautifulSoup

from apsimNGpy.tests.automate_tests import get_last_installed, get_next_version, v_next, extract_version_from_url
from apsimNGpy.tests.tester_main import send_report

email = os.environ.get('RECEIVER')


def notify_installer():

    try:
        LAST = get_last_installed()
        # logger.info("last apsim version is {}".format(LAST))
        NEXTVERSION = get_next_version(LAST)

        session = requests.Session()
        headers = {
            "User-Agent": "Mozilla/5.0"
        }

        # Step 1: Get the registration page to extract the token
        resp = session.get(
            "https://registration.apsim.info/?product=APSIM%20Next%20Generation&version",
            headers=headers
        )
        soup = BeautifulSoup(resp.text, "html.parser")

        token_input = soup.find("input", {"name": "__RequestVerificationToken"})
        token = token_input["value"]

        # Step 2: Submit the form with email and token
        post_data = {
            "email": email,
            "__RequestVerificationToken": token,
            "Sender": "LandingPage"
        }
        resp2 = session.post(
            "https://registration.apsim.info/?product=APSIM%20Next%20Generation&version",
            data=post_data,
            headers=headers
        )

        # Step 3: Parse the returned HTML for download links
        soup2 = BeautifulSoup(resp2.text, "html.parser")
        links = soup2.find_all("a")

        # Step 4: Find the correct version
        download_link = None
        vp = v_next(previous=LAST)
        for link in links:
            href = link.get("href", "")
            if f'{vp}' in href:
                print(href)
                next_v = extract_version_from_url(href)
                msg = f"""
                  A new version of apsim: {next_v} is now available.
                 """
                send_report(msg, subject="apsimNGpy version notification")

    finally:
        pass


if __name__ == "__main__":
    import schedule, time, gc

    schedule.every(1).hour.do(notify_installer)

    while True:
        schedule.run_pending()
        gc.collect()
        time.sleep(40 * 60)




from apsimNGpy.core.cal import OptimizationBase, pd
import unittest
from apsimNGpy.tests.unittests.base_unit_tests import BaseTester

data = {
    "year": [1991, 1992, 1993, 1994, 1995, 1996, 1997, 1998, 1999, 2000],
    "Yield": [7469.616, 4668.505, 555.047, 4504.0, 7820.075, 6823.517,
              3587.101, 2939.152, 6379.435, 7370.301]
}
df = pd.DataFrame(data)

print(df.columns)


class TestOptimizationBase(BaseTester):
    def _test_add_parameters(self):
        model = OptimizationBase('Maize')
        model.add_parameters(path='.Simulations.Simulation.Field.Maize.CultivarFolder.Generic.B_110',
                             commands='[Phenology].Juvenile.Target.FixedValue', values='?',
                             cultivar_manager='Sow using a variable rule', new_cultivar_name='be',
                             parameter_name='CultivarName')
        fixed_juv = '157.7'
        model._do_model_edit([fixed_juv])
        edited = model.inspect_model_parameters(model_type='Cultivar', model_name='be')[
            '[Phenology].Juvenile.Target.FixedValue']
        self.assertEqual(fixed_juv, edited, 'editing cultivar failed')

    def test_clean_data_time_year(self):
        test_df = pd.DataFrame(data)
        model = OptimizationBase('Maize')
        clean_data = model.clean_data(obs=test_df, data_table='Report', output_column='Yield', obs_time_column='year',
                                      time_step='year')
        self.assertEqual(clean_data.empty, False, msg='Cleaning with year failed')

    def _test_clean_data_time_day(self):
        test_df = df.copy(deep=True)
        test_df['day'] = [i + 1 for i in range(test_df.shape[0])]
        print(test_df['day'])
        model = OptimizationBase('Maize')
        clean_data = model.clean_data(obs=test_df, data_table='Report', output_column='Yield', obs_time_column='day',
                                      time_step='day')
        self.assertEqual(clean_data.empty, False, 'Cleaning with day failed')


if __name__ == "__main__":
    # synthetic data

    mod = OptimizationBase('Maize')
    ar = df.to_numpy(copy=True)
    df = mod.clean_data(obs=df, data_table='Report', output_column='Yield', obs_time_column='year', time_step='year')
    mod.add_parameters('.Simulations.Simulation.Field.SurfaceOrganicMatter', InitialCNR='?', InitialCPR=10)
    # example of adding cultivar
    mod.add_parameters(path='.Simulations.Simulation.Field.Maize.CultivarFolder.Generic.B_110',
                       commands='[Phenology].Juvenile.Target.FixedValue', values='?',
                       cultivar_manager='Sow using a variable rule', new_cultivar_name='be',
                       parameter_name='CultivarName')
    mod._do_model_edit([23, 156])

    mod.inspect_model_parameters(model_type='Cultivar', model_name='be')
    unittest.main()
